#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""PY4WEB - a web framework for rapid development of efficient database driven web applications"""

# Standard modules
import argparse
import cgitb
import code
import copy
import datetime
import functools
import importlib.util
import importlib.machinery
import inspect
import json
import http.client
import http.cookies
import linecache
import logging
import numbers
import os
import getpass
import pathlib
import platform
import re
import signal
import site
import shutil
import sys
import threading
import time
import traceback
import types
import urllib.parse
import uuid
import zipfile
import asyncio
from watchgod import awatch
from . import server_adapters

# Optional web servers for speed
try:
    import gunicorn
except ImportError:
    gunicorn = None

try:
    from enum import Enum
except:
    # for python < 3.4
    class Enum:
        pass


# Third party modules
import click
import bottle
import jwt  # this is PyJWT
import yatl
import pydal
import pluralize
from pydal._compat import to_native, to_bytes
import threadsafevariable

bottle.BaseRequest.MEMFILE_MAX = 16 * 1024 * 1024

__all__ = [
    "render",
    "DAL",
    "Field",
    "action",
    "request",
    "response",
    "redirect",
    "abort",
    "HTTP",
    "Session",
    "Cache",
    "Flash",
    "user_in",
    "Translator",
    "URL",
    "check_compatible",
    "required_folder",
    "wsgi",
]

PY4WEB_CMD = sys.argv[0]

REGEX_APPJSON = r"(^|\s|,)application/json(,|\s|$)"

DEFAULTS = dict(
    PY4WEB_APPS_FOLDER="apps",
    PY4WEB_SERVICE_FOLDER=".service",
    PY4WEB_SERVICE_DB_URI="sqlite://service.storage",
)

HELPERS = {name: getattr(yatl.helpers, name) for name in yatl.helpers.__all__}

ART = r"""
██████╗ ██╗   ██╗██╗  ██╗██╗    ██╗███████╗██████╗
██╔══██╗╚██╗ ██╔╝██║  ██║██║    ██║██╔════╝██╔══██╗
██████╔╝ ╚████╔╝ ███████║██║ █╗ ██║█████╗  ██████╔╝
██╔═══╝   ╚██╔╝  ╚════██║██║███╗██║██╔══╝  ██╔══██╗
██║        ██║        ██║╚███╔███╔╝███████╗██████╔╝
╚═╝        ╚═╝        ╚═╝ ╚══╝╚══╝ ╚══════╝╚═════╝
Is still experimental...
"""

Field = pydal.Field
render = yatl.render
request = bottle.request
response = bottle.response
abort = bottle.abort

os.environ.update(
    {key: value for key, value in DEFAULTS.items() if not key in os.environ}
)
os.environ["PY4WEB_PATH"] = str(pathlib.Path(__file__).resolve().parents[1])


def module2filename(module):
    filename = os.path.join(*module.split(".")[1:])
    filename = (
        os.path.join(filename, "__init__.py")
        if not filename.count(os.sep)
        else filename + ".py"
    )
    return filename


def required_folder(*parts):
    """joins the args and creates the folder if not exists"""
    path = os.path.join(*parts)
    if not os.path.exists(path):
        os.makedirs(path)
    assert os.path.isdir(path), "%s is not a folder as required" % path
    return path


########################################################################################
# fix request.fullpath for the case of domain mapping to app
# (request.url will be autofixed, since it is based on request.fullpath)
#########################################################################################
def monkey_patch_bottle():
    urljoin = urllib.parse.urljoin

    @property
    def fullpath(self):
        appname = self.environ.get("HTTP_X_PY4WEB_APPNAME", "/")
        return urljoin(self.script_name, self.path[len(appname) :])

    setattr(bottle.BaseRequest, "fullpath", fullpath)


monkey_patch_bottle()


########################################################################################
# Implement a O(1) LRU cache and memoize with expiration and monitoring (using linked list)
#########################################################################################


class Node:
    def __init__(self, key=None, value=None, t=None, m=None, prev=None, next=None):
        self.key, self.value, self.t, self.m, self.prev, self.next = (
            key,
            value,
            t,
            m,
            prev,
            next,
        )


class Cache:
    """
    O(1) caching object that remembers the 'size' most recent values
    Example:

        cache = Cache(size=1000)
        h = cache.get(filename, lambda: hash(
            open(filename).read()), 60, lambda: os.path.getmtime())

    (computes and cashes the hash of file filename but only reads the file if mtime changes and
     does not check the mtime more oftern than every 60. caches the 1000 most recent hashes)
    """

    def __init__(self, size=1000):
        self.free = size
        self.head = Node()
        self.tail = Node()
        self.head.next = self.tail
        self.tail.prev = self.head
        self.mapping = {}
        self.lock = threading.Lock()

    def get(self, key, callback, expiration=3600, monitor=None):
        """If key not stored or key has expired and monitor == None or monitor() value has changed, returns value = callback()"""
        node, t0 = self.mapping.get(key), time.time()
        with self.lock:
            if node:
                # if a node was found remove it from storage
                value, t, node.next.prev, node.prev.next = (
                    node.value,
                    node.t,
                    node.prev,
                    node.next,
                )
            else:
                self.free -= 1
        # check if something may invalidate cache
        m = monitor() if monitor else None
        # check if cache expired
        if node and node.t + expiration < t0:
            # if cache should always be invalidated or m changed
            if m is None or node.m != m:
                # ignore the value found
                node = None
        if node is None:
            value, t = callback(), t0
        # add the new node back into storage
        with self.lock:
            new_node = Node(key, value, t, m, prev=self.head, next=self.head.next)
            self.mapping[key] = self.head.next = new_node.next.prev = new_node
            if self.free < 0:
                last_node = self.tail.prev
                self.tail.prev, last_node.prev.next = last_node.prev, self.tail
                del self.mapping[last_node.key]
                self.free += 1
        return value

    def memoize(self, expiration=3600):
        def decorator(func):
            @functools.wraps(func)
            def memoized_func(*args, **kwargs):
                key = "%s:%s:%s:%s" % (func.__module__, func.__name__, args, kwargs)
                return self.get(
                    key,
                    lambda args=args, kwargs=kwargs: func(*args, **kwargs),
                    expiration=expiration,
                )

            return memoized_func

        return decorator


#########################################################################################
# A Better JSON Serializer
#########################################################################################


def objectify(obj):
    """converts the obj(ect) into a json serializable object"""
    if isinstance(obj, numbers.Integral):
        return int(obj)
    elif isinstance(obj, (numbers.Rational, numbers.Real)):
        return float(obj)
    elif isinstance(obj, (datetime.date, datetime.datetime, datetime.time)):
        return obj.isoformat()
    elif isinstance(obj, str):
        return obj
    elif hasattr(obj, "__iter__") or isinstance(obj, types.GeneratorType):
        return list(obj)
    elif hasattr(obj, "xml"):
        return obj.xml()
    elif isinstance(
        obj, Enum
    ):  # Enum class handled specially to address self reference in __dict__
        return dict(name=obj.name, value=obj.value, __class__=obj.__class__.__name__)
    elif hasattr(obj, "__dict__") and hasattr(obj, "__class__"):
        d = dict(obj.__dict__)
        d["__class__"] = obj.__class__.__name__
        return d
    return str(obj)


def dumps(obj, sort_keys=True, indent=2):
    return json.dumps(obj, default=objectify, sort_keys=sort_keys, indent=indent)


#########################################################################################
# Base Fixture (database connections, templates, sessions, and requirements are fixtures)
#########################################################################################


class Fixture:
    def on_request(self):
        pass  # called when a request arrives

    def on_error(self):
        pass  # called when a request errors

    def on_success(self, status):
        pass  # called when a request is successful

    def transform(
        self, output, shared_data=None
    ):  # transforms the output, for example to apply template
        return output


class Translator(pluralize.Translator, Fixture):
    def on_request(self):
        self.select(request.headers.get("Accept-Language", "en"))

    def on_success(self, status):
        response.headers["Content-Language"] = self.local.tag


class DAL(pydal.DAL, Fixture):

    reconnect_on_request = True

    def on_request(self):
        if self.reconnect_on_request:
            self._adapter.reconnect()
        threadsafevariable.ThreadSafeVariable.restore(ICECUBE)

    def on_error(self):
        self.rollback()

    def on_success(self, status):
        self.commit()


# make sure some variables in pydal are thread safe
def thread_safe_pydal_patch():
    Field = pydal.DAL.Field
    tsafe_attrs = ["readable", "writable", "default", "update", "requires"]
    for a in tsafe_attrs:
        setattr(Field, a, threadsafevariable.ThreadSafeVariable())

    # hack 'copy.copy' behavior, since it makes a shallow copy,
    # but ThreadSafe-attributes (see above) are class-level, so:
    # no copy -> no attr in ICECUBE for the fresh one -> gevent-error on try to access to any of ThreadSafe-attributes
    def field_copy(self):
        # to prevent infinite recursion
        # temporarily set __copy__ to None
        me = self.__class__.__copy__
        self.__class__.__copy__ = None
        clone = copy.copy(self)
        self.__class__.__copy__ = me
        for a in tsafe_attrs:
            setattr(clone, a, getattr(self, a))
        return clone

    # to avoid possible future problems
    if hasattr(Field, "__copy__"):
        raise RuntimeError("code fix required!")
    setattr(Field, "__copy__", field_copy)


thread_safe_pydal_patch()

# this global object will be used to store their state to restore it for every http request
ICECUBE = {}

#########################################################################################
# Flash Fixture
#########################################################################################


class Flash(Fixture):
    """
    flash = Flash()

    @action('index')
    @action.uses(flash)
    def index():
        flash.set('hello', _class='important')
        return dict()

    Flash messages are added to the dict and, upon redirect, carry forward
    Also notice all Flash objects share the same threading local so act as singletons
    """

    local = threading.local()

    def on_request(self):
        # when a new request arrives we look for a flash message in the cookie
        flash = request.get_cookie("py4web-flash")
        if flash:
            Flash.local.flash = json.loads(flash)
        else:
            Flash.local.flash = None

    def on_success(self, status):
        # if we redirect and have a flash message we move it to the session
        if status == 303 and Flash.local.flash:
            response.set_cookie("py4web-flash", json.dumps(Flash.local.flash), path="/")
            Flash.local.flash = None
        else:
            response.delete_cookie("py4web-flash", path="/")

    def set(self, message, _class="", sanitize=True):
        # we set a flash message
        if sanitize:
            message = yatl.sanitizer.xmlescape(message)
        Flash.local.flash = {"message": message, "class": _class}

    def transform(self, data, shared_data=None):
        # if we have a valid flash message, we inject it in the response dict
        if isinstance(data, dict):
            if not "flash" in data:
                data["flash"] = Flash.local.flash or ""
        else:
            if Flash.local.flash is not None:
                response.headers["component-flash"] = json.dumps(Flash.local.flash)
        Flash.local.flash = None
        return data


#########################################################################################
# The Template Rendered Fixture
#########################################################################################


class Template(Fixture):

    cache = Cache(100)

    def __init__(self, filename, path=None, delimiters="[[ ]]"):
        self.filename = filename
        self.path = path
        self.delimiters = delimiters

    @staticmethod
    def reader(filename):
        """Cached file reader, only reads template if it has changed"""

        def raw_read():
            with open(filename, encoding="utf8") as stream:
                return stream.read()

        return Template.cache.get(
            filename, raw_read, expiration=1, monitor=lambda: os.path.getmtime(filename)
        )

    def transform(self, output, shared_data=None):
        if not isinstance(output, dict):
            return output
        context = dict(request=request)
        context.update(HELPERS)
        context.update(URL=URL)
        if shared_data:
            context.update(shared_data.get("template_context", {}))
        context.update(output)
        context["__vars__"] = output
        app_folder = os.path.join(os.environ["PY4WEB_APPS_FOLDER"], request.app_name)
        path = self.path or os.path.join(app_folder, "templates")
        filename = os.path.join(path, self.filename)
        if not os.path.exists(filename):
            generic_filename = os.path.join(path, "generic.html")
            if os.path.exists(generic_filename):
                filename = generic_filename
        output = yatl.render(
            Template.reader(filename),
            path=path,
            context=context,
            delimiters=self.delimiters,
            reader=Template.reader,
        )
        return output


#########################################################################################
# The Session Fixture
#########################################################################################


class Session(Fixture):

    # All apps share the same default secret if not specified.
    # important for _dashboard reload
    # the actual value is loaded from a file
    SECRET = None

    def __init__(
        self,
        secret=None,
        expiration=None,
        algorithm="HS256",
        storage=None,
        same_site="Lax",
    ):
        """
        secret is the shared key used to encrypt the session (using algorithm)
        expiration is in seconds
        (optional) storage must have a get(key) and set(key,value,expiration) methods
        if not provided session is stored in jwt cookie else the jwt is stored in storage and its uuid key is stored in the cookie
        """
        # assert Session.SECRET, "Missing Session.SECRET"
        self.secret = secret or Session.SECRET
        self.expiration = expiration
        self.algorithm = algorithm
        self.local = threading.local()
        self.local.changed = False
        self.local.secure = None
        self.local.data = {}
        self.storage = storage
        self.same_site = same_site
        if isinstance(storage, Session):
            self.__prerequisites__ = [storage]
        if hasattr(storage, "__prerequisites__"):
            self.__prerequisites__ = storage.__prerequisites__

    def load(self):
        self.local.session_cookie_name = "%s_session" % request.app_name
        self.local.changed = False
        self.local.secure = request.url.startswith("https")
        self.local.data = {}
        raw_token = request.get_cookie(
            self.local.session_cookie_name
        ) or request.query.get("_session_token")
        if not raw_token and request.method in ("POST", "PUT", "DELETE"):
            raw_token = (request.forms and request.forms.get("_session_token")) or (
                request.json and request.json.get("_session_token")
            )
        if raw_token:
            token_data = to_bytes(raw_token)
            try:
                if self.storage:
                    json_data = self.storage.get(token_data)
                    if json_data:
                        self.local.data = json.loads(json_data)
                else:
                    self.local.data = jwt.decode(
                        token_data, self.secret, algorithms=[self.algorithm]
                    )
                if self.expiration is not None and self.storage is None:
                    assert self.local.data["timestamp"] > time.time() - int(
                        self.expiration
                    )
                assert self.get_data().get("secure") == self.local.secure
            except Exception:
                pass
        if not "uuid" in self.get_data():
            self.clear()

    def get_data(self):
        return getattr(self.local, "data", {})

    def save(self):
        self.local.data["timestamp"] = time.time()
        if self.storage:
            cookie_data = self.local.data["uuid"]
            self.storage.set(cookie_data, json.dumps(self.local.data), self.expiration)
        else:
            cookie_data = jwt.encode(
                self.local.data, self.secret, algorithm=self.algorithm
            )

        response.set_cookie(
            self.local.session_cookie_name,
            to_native(cookie_data),
            path="/",
            secure=self.local.secure,
            same_site=self.same_site,
        )

    def get(self, key, default=None):
        return self.get_data().get(key, default)

    def __getitem__(self, key):
        return self.get_data()[key]

    def __delitem__(self, key):
        if key in self.get_data():
            self.local.changed = True
            del self.local.data[key]

    def __setitem__(self, key, value):
        self.local.changed = True
        self.local.data[key] = value

    def keys(self):
        return self.get_data().keys()

    def __iter__(self):
        for item in self.get_data().items():
            yield item

    def clear(self):
        self.local.changed = True
        self.local.data.clear()
        self.local.data["uuid"] = str(uuid.uuid1())
        self.local.data["secure"] = self.local.secure

    def on_request(self):
        self.load()

    def on_error(self):
        if self.local.changed:
            self.save()

    def on_success(self, status):
        if self.local.changed:
            self.save()


#########################################################################################
# The URL Helper
#########################################################################################


def URL(
    *parts,
    vars=None,
    hash=None,
    scheme=False,
    signer=None,
    use_appname=None,
    static_version=None,
):
    """
    Examples:
    URL('a','b',vars=dict(x=1),hash='y')       -> /{script_name?}/{app_name?}/a/b?x=1#y
    URL('a','b',vars=dict(x=1),scheme=None)    -> //{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),scheme=True)    -> http://{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),scheme='https') -> https://{domain}/{script_name?}/{app_name?}/a/b?x=1
    URL('a','b',vars=dict(x=1),use_appname=False) -> /{script_name?}/a/b?x=1
    """
    if use_appname is None:
        # force use_appname on domain-unmapped apps
        use_appname = not request.environ.get("HTTP_X_PY4WEB_APPNAME")
    if use_appname:
        # app_name is not set by py4web shell
        app_name = getattr(request, "app_name", None)
    has_appname = use_appname and app_name
    script_name = (
        request.environ.get("SCRIPT_NAME", "")
        or request.environ.get("HTTP_X_SCRIPT_NAME", "")
    ).rstrip("/")
    if parts and parts[0].startswith("/"):
        prefix = ""
    elif has_appname and app_name != "_default":
        prefix = "%s/%s/" % (script_name, app_name)
    else:
        prefix = "%s/" % script_name
    broken_parts = []
    for part in parts:
        broken_parts += str(part).rstrip("/").split("/")
    if static_version != "" and broken_parts and broken_parts[0] == "static":
        if not static_version:
            # try to retrieve from __init__.py
            app_module = "apps.%s" % app_name if has_appname else "apps"
            try:
                static_version = getattr(
                    sys.modules[app_module], "__static_version__", None
                )
            except KeyError:
                static_version = None
        if static_version:
            broken_parts.insert(1, "_" + static_version)

    url = prefix + "/".join(map(urllib.parse.quote, broken_parts))
    # Signs the URL if required.  Copy vars into urlvars not to modify it.
    urlvars = dict(vars) if vars else {}
    if signer:
        # Note that we need to sign the non-urlencoded URL, since
        # at verification time, it will be already URLdecoded.
        signer.sign(prefix + "/".join(broken_parts), urlvars)
    if urlvars:
        url += "?" + "&".join(
            "%s=%s" % (k, urllib.parse.quote(str(v))) for k, v in urlvars.items()
        )
    if hash:
        url += "#%s" % hash
    if not scheme is False:
        original_url = request.environ.get("HTTP_ORIGIN") or request.url
        orig_scheme, _, domain = original_url.split("/", 3)[:3]
        if scheme is True:
            scheme = orig_scheme
        elif scheme is None:
            scheme = ""
        else:
            scheme += ":"
        url = "%s//%s%s" % (scheme, domain, url)
    return url


#########################################################################################
# The Action Decorator
#########################################################################################


class HTTP(BaseException):
    class Type:
        success = "success"
        error = "error"

    """Our HTTP exception does not delete cookies and headers like the bottle.HTTPResponse does;
    since it is considered a success, not a failure"""

    def __init__(self, status, type=Type.success):
        self.status = status
        self.type = type


def redirect(location):
    """our redirect does not delete cookies and headers like bottle.HTTPResponse does;
    it is considered a success, not failure"""
    response.set_header("Location", location)
    raise HTTP(303)


class action:
    """@action(...) is a decorator for functions to be exposed as actions"""

    current = threading.local()
    registered = set()
    app_name = "_default"

    def __init__(self, path, **kwargs):
        self.path = path
        self.kwargs = kwargs

    @staticmethod
    def uses(*fixtures_in):
        """Find all fixtures, including dependencies, topologically sorted"""
        fixtures = []
        reversed_fixtures = []
        stack = list(fixtures_in)
        while stack:
            fixture = stack.pop()
            reversed_fixtures.append(fixture)
            stack.extend(getattr(fixture, "__prerequisites__", ()))
        for fixture in reversed(reversed_fixtures):
            if isinstance(fixture, str):
                fixture = Template(fixture)
            if not fixture in fixtures:
                fixtures.append(fixture)

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # data shared by all fixtures in the pipeline for each request
                shared_data = {"template_context": {}}
                try:
                    [obj.on_request() for obj in fixtures]
                    ret = func(*args, **kwargs)
                    for obj in fixtures:
                        ret = obj.transform(ret, shared_data)
                    [obj.on_success(200) for obj in fixtures]
                    return ret
                except HTTP as http:
                    if http.type == http.Type.success:
                        [obj.on_success(http.status) for obj in fixtures]
                    else:
                        [obj.on_error() for obj in fixtures]
                        # it should be [obj.on_error(status) for obj in fixtures]
                        # but it breaks users fixtures
                        # `def on_error(status = None):` - cost nothing, but we have  `def on_error():`
                    raise
                except Exception:
                    [obj.on_error() for obj in fixtures]
                    raise

            return wrapper

        return decorator

    @staticmethod
    def requires(*requirements):
        """Enforces requirements or calls bottle.abort(401)"""

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                for requirement in requirements:
                    if not requirement():
                        bottle.abort(401)
                return func(*args, **kwargs)

            return wrapper

        return decorator

    @staticmethod
    def catch_errors(app_name, func):
        """Catches and logs errors in an action; also sets request.app_name"""

        @functools.wraps(func)
        def wrapper(*func_args, **func_kwargs):
            try:
                request.app_name = app_name
                ret = func(*func_args, **func_kwargs)
                if isinstance(ret, dict):
                    response.headers["Content-Type"] = "application/json"
                    ret = dumps(ret)
                return ret
            except HTTP as http:
                response.status = http.status
                return getattr(http, "body", "")
            except bottle.HTTPResponse:
                raise
            except Exception:
                logging.error(traceback.format_exc())
                try:
                    ticket = ErrorStorage().log(request.app_name, get_error_snapshot())
                except Exception:
                    logging.error(traceback.format_exc())
                    ticket = "unknown"
                raise bottle.HTTPResponse(
                    body=error_page(
                        500, button_text=ticket, href="/_dashboard/ticket/" + ticket
                    ),
                    status=500,
                )

        return wrapper

    def __call__(self, func):
        """Building the decorator"""
        trailing = "<:re:/?>"
        app_name = action.app_name
        base_path = "" if app_name == "_default" else "/%s" % app_name
        path = (base_path + "/" + self.path).rstrip("/")
        if not func in self.registered:
            func = action.catch_errors(app_name, func)
        func = bottle.route(path + trailing, **self.kwargs)(func)
        if path.endswith("/index"):  # /index is always optional
            func = bottle.route(path[:-6] + trailing, **self.kwargs)(func)
        self.registered.add(func)
        return func


def user_in(session):
    def requirement():
        session.on_request()
        return session.get("user", None) is not None

    return requirement


#########################################################################################
# Monkey Patch: Cookies
#########################################################################################

http.cookies.Morsel._reserved["same-site"] = "SameSite"

#########################################################################################
# Monkey Patch: ssl bug for gevent
#########################################################################################

__ssl__ = __import__("ssl")
_ssl = getattr(__ssl__, "_ssl") or getattr(__ssl__, "_ssl2")


def new_sslwrap(
    sock,
    server_side=False,
    keyfile=None,
    certfile=None,
    cert_reqs=__ssl__.CERT_NONE,
    ssl_version=__ssl__.PROTOCOL_SSLv23,
    ca_certs=None,
    ciphers=None,
):
    context = __ssl__.SSLContext(ssl_version)
    context.verify_mode = cert_reqs or __ssl__.CERT_NONE
    if ca_certs:
        context.load_verify_locations(ca_certs)
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    if ciphers:
        context.set_ciphers(ciphers)
    caller_self = inspect.currentframe().f_back.f_locals["self"]
    return context._wrap_socket(sock, server_side=server_side, ssl_sock=caller_self)


#########################################################################################
# Error Handling
#########################################################################################


def get_error_snapshot(depth=5):
    """Return a dict describing a given traceback (based on cgitb.text)."""

    etype, evalue, etb = sys.exc_info()
    if isinstance(etype, type):
        etype = etype.__name__

    data = {}
    data["timestamp"] = datetime.datetime.utcnow().isoformat()
    data["python_version"] = sys.version
    platform_keys = [
        "machine",
        "node",
        "platform",
        "processor",
        "python_branch",
        "python_build",
        "python_compiler",
        "python_implementation",
        "python_revision",
        "python_version",
        "python_version_tuple",
        "release",
        "system",
        "uname",
        "version",
    ]

    data["platform_info"] = {key: getattr(platform, key)() for key in platform_keys}
    data["os_environ"] = {key: str(value) for key, value in os.environ.items()}
    data["traceback"] = traceback.format_exc()
    data["exception_type"] = str(etype)
    data["exception_value"] = str(evalue)

    # Loopover the stack frames
    items = inspect.getinnerframes(etb, depth)
    del etb  # Prevent circular references that would cause memory leaks
    data["stackframes"] = stackframes = []

    for frame, file, lnum, func, lines, idx in items:
        file = file and os.path.abspath(file) or "?"
        args, varargs, varkw, locals = inspect.getargvalues(frame)
        # Basic frame information
        f = {"file": file, "func": func, "lnum": lnum}
        f["code"] = lines
        # FIXME: disable this for now until we understand why this goes into infinite loop
        if False:
            line_vars = cgitb.scanvars(
                lambda: linecache.getline(file, lnum), frame, locals
            )
            # Dump local variables (referenced in current line only)
            f["vars"] = {
                key: repr(value)
                for key, value in locals.items()
                if not key.startswith("__")
            }
        stackframes.append(f)

    return data


class ErrorStorage:
    def __init__(self):
        uri = os.environ["PY4WEB_SERVICE_DB_URI"]
        folder = os.environ["PY4WEB_SERVICE_FOLDER"]
        self.db = DAL(uri, folder=folder)
        self.db.define_table(
            "py4web_error",
            Field("uuid"),
            Field("app_name"),
            Field("method"),
            Field("path", "string"),
            Field("timestamp", "datetime"),
            Field("client_ip", "string"),
            Field("error", "string"),
            Field("snapshot", "json"),
        )
        self.db.commit()

    def log(self, app_name, error_snapshot):
        ticket_uuid = str(uuid.uuid4())
        try:
            id = self.db.py4web_error.insert(
                uuid=ticket_uuid,
                app_name=app_name,
                method=request.method,
                path=request.path,
                timestamp=datetime.datetime.utcnow(),
                client_ip=request.environ.get("REMOTE_ADDR"),
                error=error_snapshot["exception_value"],
                snapshot=error_snapshot,
            )
            self.db.commit()
            return ticket_uuid
        except Exception:
            self.db.rollback()
            return "internal-error"

    def get(self, ticket_uuid=None):
        db = self.db
        if ticket_uuid:
            query, orderby = db.py4web_error.uuid == ticket_uuid, None
            rows = db(query).select(orderby=orderby, limitby=(0, 1)).as_list()
        else:
            orderby = ~db.py4web_error.timestamp
            groupby = db.py4web_error.path | db.py4web_error.error
            query = (
                db.py4web_error.timestamp
                > datetime.datetime.now() - datetime.timedelta(days=7)
            )
            fields = [field for field in db.py4web_error if not field.type == "json"]
            fields.append(db.py4web_error.id.count())
            list_rows = (
                db(query).select(*fields, orderby=orderby, groupby=groupby).as_list()
            )
            rows = []
            for item in list_rows:
                row = item["py4web_error"]
                row["count"] = item["_extra"][str(db.py4web_error.id.count())]
                rows.append(row)
        return rows if not ticket_uuid else rows[0] if rows else None

    def clear(self):
        db = self.db
        db(db.py4web_error).delete()
        self.db.commit()


#########################################################################################
# Loading &  Reloading Logic
#########################################################################################


class Reloader:

    ROUTES = []
    MODULES = {}
    ERRORS = {}

    @staticmethod
    def install_reloader_hook():
        # used by watcher
        def hook(*a, **k):
            app_name = request.path.split("/")[1]
            if app_name in DIRTY_APPS:
                Reloader.import_app(app_name)
                del DIRTY_APPS[app_name]
            ## APP_WATCH tasks, if used by any app
            try_app_watch_tasks()

        bottle.default_app().add_hook("before_request", hook)

    @staticmethod
    def clear_routes(app_name=None):
        app = bottle.default_app()
        routes = app.routes[:]
        app.routes.clear()
        app.router = bottle.Router()
        if app_name:
            for route in routes:
                if route.rule.rstrip("<:re:/?>")[1:].split("/")[0] != app_name:
                    app.add_route(route)

    @staticmethod
    def import_apps():
        """Import or reimport modules and exposed static files"""
        Reloader.clear_routes()
        folder = os.environ["PY4WEB_APPS_FOLDER"]
        # if first time reload dummy top module
        if not Reloader.MODULES:
            path = os.path.join(folder, "__init__.py")
            loader = importlib.machinery.SourceFileLoader("apps", path)
            loader.load_module()
        # Then load all the apps as submodules
        for app_name in os.listdir(folder):
            Reloader.import_app(app_name, clear_before_import=False)

    @staticmethod
    def import_app(app_name, clear_before_import=True):
        if clear_before_import:
            Reloader.clear_routes(app_name)
        folder = os.environ["PY4WEB_APPS_FOLDER"]
        path = os.path.join(folder, app_name)
        init = os.path.join(path, "__init__.py")

        if os.path.isdir(path) and not path.endswith("__") and os.path.exists(init):

            action.app_name = app_name
            module_name = "apps.%s" % app_name

            def clear_modules():
                # all files/submodules
                names = [
                    name
                    for name in sys.modules
                    if (name + ".").startswith(module_name + ".")
                ]
                for name in names:
                    del sys.modules[name]

            try:
                module = Reloader.MODULES.get(app_name)
                if not module:
                    click.echo("[ ] loading %s ..." % app_name)
                else:
                    click.echo("[ ] reloading %s ..." % app_name)
                    # forget the module
                    del Reloader.MODULES[app_name]
                    clear_modules()
                module = importlib.machinery.SourceFileLoader(
                    module_name, init
                ).load_module()
                click.secho("\x1b[A[X] loaded %s       " % app_name, fg="green")
                Reloader.MODULES[app_name] = module
                Reloader.ERRORS[app_name] = None
            except:
                tb = traceback.format_exc()
                print(tb)
                click.secho(
                    "\x1b[A[FAILED] loading %s       \n%s\n" % (app_name, tb),
                    fg="red",
                )
                Reloader.ERRORS[app_name] = tb
                # clear all files/submodules if the loading fails
                clear_modules()
                return None

        # Expose static files with support for static asset management
        static_folder = os.path.join(path, "static")

        if os.path.exists(static_folder):
            app_name = path.split(os.path.sep)[-1]
            prefix = "" if app_name == "_default" else ("/%s" % app_name)

            @bottle.route(prefix + "/static/<filename:path>")
            @bottle.route(
                prefix + "/static/_<version:re:\\d+\\.\\d+\\.\\d+>/<filename:path>"
            )
            def server_static(filename, static_folder=static_folder, version=None):
                response.headers.setdefault("Pragma", "cache")
                response.headers.setdefault("Cache-Control", "private")
                return bottle.static_file(filename, root=static_folder)

        # Register routes list
        app = bottle.default_app()
        routes = []
        for route in app.routes:
            func = route.callback
            rule = route.rule
            # remove optional trailing / from rule
            if rule.endswith("<:re:/?>"):
                rule = rule[:-8]
            routes.append(
                {
                    "rule": rule,
                    "method": route.method,
                    "filename": module2filename(func.__module__),
                    "action": func.__name__,
                }
            )
        Reloader.ROUTES = sorted(routes, key=lambda item: item["rule"])
        ICECUBE.update(threadsafevariable.ThreadSafeVariable.freeze())


#########################################################################################
# Web Server and Reload Logic: Error Handling
#########################################################################################


def error_page(code, button_text=None, href="#", color=None, message=None):
    message = http.client.responses[code].upper() if message is None else message
    color = (
        {"4": "#F44336", "5": "#607D8B"}.get(str(code)[0], "#2196F3")
        if not color
        else color
    )
    context = dict(
        code=code, message=message, button_text=button_text, href=href, color=color
    )
    # if client accepts 'application/json' - return json
    if re.search(REGEX_APPJSON, request.headers.get("accept", "")):
        response.status = code
        return json.dumps(context)
    # else - return html error-page
    return yatl.render(
        '<html><head><style>body{color:white;text-align: center;background-color:[[=color]];font-family:serif} h1{font-size:6em;margin:16vh 0 8vh 0} h2{font-size:2em;margin:8vh 0} a{color:white;text-decoration:none;font-weight:bold;padding:10px 10px;border-radius:10px;border:2px solid #fff;transition: all .5s ease} a:hover{background:rgba(0,0,0,0.1);padding:10px 30px}</style></head><body><h1>[[=code]]</h1><h2>[[=message]]</h2>[[if button_text:]]<a href="[[=href]]">[[=button_text]]</a>[[pass]]</body></html>',
        context=context,
        delimiters="[[ ]]",
    )


@bottle.error(404)
def error404(error):
    guess_app_name = (
        "index"
        if request.environ.get("HTTP_X_PY4WEB_APPNAME")
        else request.path.split("/")[1]
    )
    if guess_app_name == "index":
        href = "/"
    else:
        href = "/" + guess_app_name
    script_name = (
        request.environ.get("SCRIPT_NAME", "")
        or request.environ.get("HTTP_X_SCRIPT_NAME", "")
    ).rstrip("/")
    if script_name:
        href = script_name + href
    return error_page(404, button_text=guess_app_name, href=href)


#########################################################################################
# Web Server and Reload Logic: Operations
#########################################################################################

DIRTY_APPS = dict()  #  apps that need to be reloaded (lazy watching)

from inspect import stack
from collections import OrderedDict

APP_WATCH = {"files": dict(), "handlers": OrderedDict(), "tasks": dict()}

""" Decorator that binds a func as an watchdog handler of non-'.py' files.
Paths to files must be relative to app, w/o app name(folder).

@app_watch_handler(['static/sass/all.sass', 'static/sass/main.sass'])
def sass_compile(changed_files):
    print(changed_files); # paths of files that changed, for info
    sass.compile()
"""


def app_watch_handler(watched_app_subpaths):
    invoker = pathlib.Path(stack()[1].filename)
    apps_path = pathlib.Path(os.environ["PY4WEB_APPS_FOLDER"])
    app = invoker.relative_to(os.environ["PY4WEB_APPS_FOLDER"]).parts[0]

    def decorator(func):
        handler = "{}.{}".format(func.__module__, func.__name__)
        APP_WATCH["handlers"][handler] = func
        for subpath in watched_app_subpaths:
            app_path = apps_path.joinpath(app, subpath).as_posix()
            if not app_path in APP_WATCH["files"]:
                APP_WATCH["files"][app_path] = []
            APP_WATCH["files"][app_path].append(handler)
        return func

    return decorator


def try_app_watch_tasks():
    if APP_WATCH["tasks"]:
        tried_tasks = []
        for handler in APP_WATCH["tasks"]:
            changed_files_dict = APP_WATCH["tasks"][handler]
            try:
                APP_WATCH["handlers"][handler](changed_files_dict.keys())
                tried_tasks.append(handler)
            except Exception:
                logging.error(traceback.format_exc())
        ## remove executed tasks from register
        for handler in tried_tasks:
            del APP_WATCH["tasks"][handler]


def watch(apps_folder, server_config, mode="sync"):
    def watch_folder_event_loop(apps_folder):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(watch_folder(apps_folder))

    async def watch_folder(apps_folder):
        click.echo(
            "watching (%s-mode) python file changes in: %s" % (mode, apps_folder)
        )
        async for changes in awatch(os.path.join(apps_folder)):
            apps = []
            for subpath in [pathlib.Path(pair[1]) for pair in changes]:
                name = subpath.relative_to(apps_folder).parts[0]
                if subpath.suffix == ".py":
                    apps.append(name)
                ## manage `app_watch_handler` decorators
                elif subpath.as_posix() in APP_WATCH["files"]:
                    handlers = APP_WATCH["files"][subpath.as_posix()]
                    for handler in handlers:
                        if not handler in APP_WATCH["tasks"]:
                            APP_WATCH["tasks"][handler] = {}
                        APP_WATCH["tasks"][handler][subpath.as_posix()] = True

            for name in apps:
                if mode == "lazy":
                    DIRTY_APPS[name] = True
                else:
                    Reloader.import_app(name)
            ## in 'lazy' mode it's done in bottle's 'before_request' hook
            if mode != "lazy":
                try_app_watch_tasks()

    if server_config["number_workers"] > 1:
        click.echo("--watch option has no effect in multi-process environment \n")
        return
    elif server_config["server"].startswith(("wsgiref", "waitress", "rocket")):
        # these servers block the main thread so we open a new thread for the file watcher
        threading.Thread(
            target=watch_folder_event_loop, args=(apps_folder,), daemon=True
        ).start()
    elif server_config["server"] == "tornado":
        # tornado delegate to asyncio so we add a future into the event loop
        asyncio.ensure_future(watch_folder(apps_folder))
    elif server_config["server"].startswith("gevent"):
        watch_folder_event_loop(apps_folder)
    else:
        # should never happen
        return

    if mode == "lazy":
        Reloader.install_reloader_hook()


def start_server(kwargs, ctrl_c_orig):
    host = kwargs["host"]
    port = int(kwargs["port"])
    apps_folder = kwargs["apps_folder"]
    number_workers = kwargs["number_workers"]
    params = dict(host=host, port=port, reloader=False)
    server_config = dict(
        platform=platform.system().lower(),
        server=None if kwargs["server"] == "default" else kwargs["server"],
        number_workers=number_workers,
    )

    if server_config["server"]:
        for e in ("rocket", "Twisted"):
            if e in server_config["server"]:
                signal.signal(signal.SIGINT, ctrl_c_orig)
                break

    if not server_config["server"]:
        if server_config["platform"] == "windows":
            server_config["server"] = "tornado"
            if sys.version_info >= (
                3,
                8,
            ):  # see  https://bugs.python.org/issue37373 FIX: tornado/py3.8 on windows
                asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        elif number_workers <= 1:
            server_config["server"] = "tornado"
        else:
            if not gunicorn:
                logging.error("gunicorn not installed")
                return
            server_config["server"] = "gunicorn"
    params["server"] = server_config["server"]
    if params["server"] in server_adapters.__all__:
        params["server"] = getattr(server_adapters, params["server"])()
    if number_workers > 1:
        params["workers"] = number_workers
    if server_config["server"] == "gunicorn":
        sys.argv[:] = sys.argv[:1]  # else break gunicorn
    if kwargs["ssl_cert"] is not None:
        params["certfile"] = kwargs["ssl_cert"]
        params["keyfile"] = kwargs["ssl_key"]

    if server_config["server"] == "gevent":
        if not hasattr(_ssl, "sslwrap"):
            _ssl.sslwrap = new_sslwrap

    if kwargs["watch"] != "off":
        watch(apps_folder, server_config, kwargs["watch"])
    bottle.run(**params)


def check_compatible(version):
    """To be called by apps to check if module version is compatible with py4web requirements"""
    from . import __version__

    return tuple(map(int, __version__.split("."))) >= tuple(
        map(int, version.split("."))
    )


#########################################################################################
# WSGI Adapter
#########################################################################################


class MetaPathRouter:
    """
    Instances of this class makes alias for a package name,
    in other words instruct the import system to route request
    for a package alias, i.e.:

        MetaPathRouter("pkg", "pkg_alias")
        import pkg_alias.sub

    works as

        import pkg.sub

    author: Paolo Pastori
    """

    def __init__(self, pkg, pkg_alias="apps"):
        assert pkg_alias
        assert pkg
        if pkg != pkg_alias:
            self.pkg_alias = pkg_alias
            self.pkg = pkg
            # register as path finder
            sys.meta_path.append(self)

    def find_spec(self, fullname, path=None, target=None):
        if fullname == self.pkg_alias and path is None:
            spec = importlib.util.find_spec(self.pkg)
            if spec:
                spec.name = fullname
                spec.loader = importlib.machinery.SourceFileLoader(
                    fullname, spec.origin
                )
                return spec


def install_args(kwargs, reinstall_apps=False):
    # always convert apps_folder to an absolute path
    apps_folder = kwargs["apps_folder"] = os.path.abspath(kwargs["apps_folder"])
    kwargs["service_folder"] = os.path.join(
        kwargs["apps_folder"], DEFAULTS["PY4WEB_SERVICE_FOLDER"]
    )
    kwargs["service_db_uri"] = DEFAULTS["PY4WEB_SERVICE_DB_URI"]
    for key, val in kwargs.items():
        os.environ["PY4WEB_" + key.upper()] = str(val)
    yes2 = yes = kwargs.get("yes", False)
    # If the apps folder does not exist create it and populate it
    if not os.path.exists(apps_folder):
        if yes or click.confirm("Create missing folder %s?" % apps_folder):
            os.makedirs(apps_folder)
            yes2 = True
        else:
            click.echo("Command aborted")
            sys.exit(0)
    init_py = os.path.join(apps_folder, "__init__.py")
    if not os.path.exists(init_py):
        if yes2 or click.confirm("Create missing init file %s?" % init_py):
            with open(init_py, "wb"):
                pass
        else:
            click.echo("Command aborted")
            sys.exit(0)
    # ensure that "import apps.someapp" works
    apps_folder_parent, apps_folder_name = os.path.split(apps_folder)
    if not apps_folder_parent in sys.path:
        sys.path.insert(0, apps_folder_parent)
    if apps_folder_name != "apps":
        MetaPathRouter(apps_folder_name)

    # Reinstall apps from zipped ones in assets
    if reinstall_apps:
        assets_dir = os.path.join(os.path.dirname(__file__), "assets")
        if os.path.exists(assets_dir):
            apps = os.listdir(assets_dir)
            for filename in apps:
                zip_filename = os.path.join(assets_dir, filename)
                # These filenames do not necessarily exist if one has
                # downloaded from source and deleted them.
                app_name = filename.split(".")[-2]
                target_dir = os.path.join(apps_folder, app_name)
                if not os.path.exists(target_dir):
                    if yes or click.confirm("Create app %s?" % app_name):
                        click.echo("[ ] Unzipping app %s" % filename)
                        with zipfile.ZipFile(zip_filename, "r") as zip_file:
                            os.makedirs(target_dir)
                            zip_file.extractall(target_dir)
                            click.echo("\x1b[A[X]")

    if not os.path.exists(kwargs["service_folder"]):
        os.mkdir(kwargs["service_folder"])
    session_secret_filename = os.path.join(kwargs["service_folder"], "session.secret")
    if not os.path.exists(session_secret_filename):
        with open(session_secret_filename, "w") as fp:
            fp.write(str(uuid.uuid4()))

    with open(session_secret_filename) as fp:
        Session.SECRET = fp.read()


def wsgi(**kwargs):
    """Initializes everything, loads apps, returns the wsgi app"""
    install_args(kwargs)
    Reloader.import_apps()
    return bottle.default_app()


#########################################################################################
# CLI
#########################################################################################


def keyboardInterruptHandler(signal, frame):
    """Catch interrupts like Ctrl-C"""
    click.echo(
        "KeyboardInterrupt (ID: {}) has been caught. Cleaning up...".format(signal)
    )
    sys.exit(0)


@click.group(
    context_settings=dict(help_option_names=["-h", "-help", "--help"]),
    help='%s\n\nType "%s COMMAND -h" for available options on commands'
    % (__doc__, PY4WEB_CMD),
)
def cli():
    pass


@cli.command()
@click.option(
    "-a", "--all", is_flag=True, default=False, help="List version of all modules"
)
def version(all):
    """Show versions and exit"""
    from . import __version__

    click.echo("py4web: %s" % __version__)
    if all:
        click.echo("system: %s" % platform.platform())
        click.echo("python: %s" % sys.version.replace("\n", " "))
        for name in sorted(sys.modules):
            if hasattr(sys.modules[name], "__version__"):
                click.echo("%s: %s" % (name, sys.modules[name].__version__))


@cli.command()
@click.argument("apps_folder")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
def setup(**kwargs):
    """Setup new apps folder or reinstall it"""
    install_args(kwargs, reinstall_apps=True)


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
def shell(**kwargs):
    """Open a python shell with apps_folder's parent added to the path"""
    install_args(kwargs)
    code.interact(local=dict(globals(), **locals()))


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.argument("func")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option(
    "--args",
    default="{}",
    help="Arguments passed to the program/function",
    show_default=True,
)
def call(apps_folder, func, yes, args):
    """Call a function inside apps_folder"""
    kwargs = json.loads(args)
    install_args(dict(apps_folder=apps_folder, yes=yes))
    apps_folder_name = os.path.basename(os.environ["PY4WEB_APPS_FOLDER"])
    module, name = ("%s.%s" % (apps_folder_name, func)).rsplit(".", 1)
    env = {}
    exec("from %s import %s" % (module, name), {}, env)
    env[name](**kwargs)


@cli.command(name="set_password")
@click.option(
    "--password",
    prompt=True,
    confirmation_prompt=True,
    hide_input=True,
    help="Password value (asked if missing)",
)
@click.option(
    "-p",
    "--password_file",
    default="password.txt",
    help="File for the encrypted password",
    show_default=True,
)
def set_password(password, password_file):
    """Set administrator's password for the Dashboard"""
    click.echo('Storing the hashed password in file "%s"\n' % password_file)
    with open(password_file, "w") as fp:
        fp.write(str(pydal.validators.CRYPT()(password)[0]))


@cli.command(name="new_app")
@click.argument("apps_folder")
@click.argument("app_name")
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option(
    "-s",
    "--scaffold_zip",
    default=None,
    help="Path to the zip with the scaffolding app",
    show_default=False,
)
def new_app(apps_folder, app_name, yes, scaffold_zip):
    """Create a new app copying the scaffolding one"""
    install_args(dict(apps_folder=apps_folder, yes=yes))
    source = scaffold_zip or os.path.join(
        os.path.dirname(__file__), "assets", "py4web.app._scaffold.zip"
    )
    target_dir = os.path.join(os.environ["PY4WEB_APPS_FOLDER"], app_name)
    if not os.path.exists(source):
        click.echo("Source app %s does not exists" % source)
        sys.exit(1)
    elif os.path.exists(target_dir):
        click.echo("Target folder %s already exists" % target_dir)
        sys.exit(1)
    else:
        zfile = zipfile.ZipFile(source, "r")
        zfile.extractall(target_dir)
        zfile.close()


@cli.command()
@click.argument("apps_folder", type=click.Path(exists=True))
@click.option(
    "-Y",
    "--yes",
    is_flag=True,
    default=False,
    help="No prompt, assume yes to questions",
    show_default=True,
)
@click.option("-H", "--host", default="127.0.0.1", help="Host name", show_default=True)
@click.option(
    "-P", "--port", default=8000, type=int, help="Port number", show_default=True
)
@click.option(
    "-p",
    "--password_file",
    default="password.txt",
    help="File for the encrypted password",
    show_default=True,
)
@click.option(
    "-s",
    "--server",
    default="default",
    type=click.Choice(
        ["default", "wsgiref", "tornado", "gunicorn", "gevent", "waitress"]
        + server_adapters.__all__
    ),
    help="server to use",
    show_default=True,
)
@click.option(
    "-w",
    "--number_workers",
    default=0,
    type=int,
    help="Number of workers",
    show_default=True,
)
@click.option(
    "-d",
    "--dashboard_mode",
    default="full",
    help="Dashboard mode: demo, readonly, full, none",
    show_default=True,
)
@click.option(
    "--watch",
    default="off",
    type=click.Choice(["off", "sync", "lazy"]),
    help="Watch python changes and reload apps automatically, modes: off, sync, lazy",
    show_default=True,
)
@click.option(
    "--ssl_cert", type=click.Path(exists=True), help="SSL certificate file for HTTPS"
)
@click.option("--ssl_key", type=click.Path(exists=True), help="SSL key file for HTTPS")
def run(**kwargs):
    """Run all the applications on apps_folder"""
    install_args(kwargs)

    from py4web import __version__

    click.secho(ART, fg="blue")
    click.echo("Py4web: %s on Python %s\n\n" % (__version__, sys.version))

    # If we know where the password is stored, read it, otherwise ask for one
    if os.path.exists(os.path.join(os.environ["PY4WEB_APPS_FOLDER"], "_dashboard")):
        if kwargs["dashboard_mode"] not in ("demo", "none") and not os.path.exists(
            kwargs["password_file"]
        ):
            click.echo(
                'You have not set a dashboard password. Run "%s set_password" to do so.'
                % PY4WEB_CMD
            )
        else:
            click.echo(
                "Dashboard is at: http://%s:%s/_dashboard"
                % (kwargs["host"], kwargs["port"])
            )

    # Catch interrupts like Ctrl-C
    orig_ctrl_c_handler = signal.getsignal(signal.SIGINT)
    signal.signal(signal.SIGINT, keyboardInterruptHandler)

    # Start
    Reloader.import_apps()
    start_server(kwargs, orig_ctrl_c_handler)


if __name__ == "__main__":
    cli()



class CAS(SSO):

    name = "cas"
    maps = {}

    ### methods that must be overwritten

    def __init__(self, cas_provider=None, cas_domains=None):
        self.parameters = dict(
            cas_provider = cas_providers,
            cas_domains = cas_domains
        )
        self.cas_actions = dict(

        )
        
    def get_login_url(self):
        """returns the url for login"""
        return ""

    def handle_request(self, auth, path, get_vars, post_vars):
        if path == "login":
            redirect(self.get_login_url())
        elif path == "callback":
            self._handle_callback(auth, get_vars)
        else:
            abort(404)

    def callback(self, get_vars):
        return {}




elif args[0] == 'cas' and not self.settings.cas_provider:
            if args(1) == self.settings.cas_actions['login']:
                return self.cas_login(version=2)
            elif args(1) == self.settings.cas_actions['validate']:
                return self.cas_validate(version=1)
            elif args(1) == self.settings.cas_actions['servicevalidate']:
                return self.cas_validate(version=2, proxy=False)
            elif args(1) == self.settings.cas_actions['proxyvalidate']:
                return self.cas_validate(version=2, proxy=True)
            elif (args(1) == 'p3'
                  and args(2) == self.settings.cas_actions['servicevalidate']):
                return self.cas_validate(version=3, proxy=False)
            elif (args(1) == 'p3'
                  and args(2) == self.settings.cas_actions['proxyvalidate']):
                return self.cas_validate(version=3, proxy=True)
            elif args(1) == self.settings.cas_actions['logout']:
                return self.logout(next=request.vars.service or DEFAULT)

            
                
                db.define_table(
                    settings.table_cas_name,
                    Field('user_id', reference_table_user, default=None,
                          label=self.messages.label_user_id),
                    Field('created_on', 'datetime', default=now),
                    Field('service', requires=IS_URL()),
                    Field('ticket'),
                    Field('renew', 'boolean', default=False),
                    *settings.extra_fields.get(settings.table_cas_name, []),
                    **dict(
                        migrate=self._get_migrate(
                            settings.table_cas_name, migrate),
                        fake_migrate=fake_migrate))

        if settings.cas_provider:  # THIS IS NOT LAZY
            settings.actions_disabled = \
                ['profile', 'register', 'change_password',
                 'request_reset_password', 'retrieve_username']
            from gluon.contrib.login_methods.cas_auth import CasAuth
            maps = settings.cas_maps
            if not maps:
                table_user = self.table_user()
                maps = dict((name, lambda v, n=name: v.get(n, None)) for name in
                            table_user.fields if name != 'id'
                            and table_user[name].readable)
                maps['registration_id'] = \
                    lambda v, p=settings.cas_provider: '%s/%s' % (p, v['user'])
            actions = [settings.cas_actions['login'],
                       settings.cas_actions['servicevalidate'],
                       settings.cas_actions['logout']]
            settings.login_form = CasAuth(
                casversion=2,
                urlbase=settings.cas_provider,
                actions=actions,
                maps=maps)
        return self

    def cas_login(self,
                  next=DEFAULT,
                  onvalidation=DEFAULT,
                  onaccept=DEFAULT,
                  log=DEFAULT,
                  version=2,
                  ):
        request = current.request
        response = current.response
        session = current.session
        db, table = self.db, self.table_cas()
        session._cas_service = request.vars.service or session._cas_service
        if request.env.http_host not in self.settings.cas_domains or \
                not session._cas_service:
            raise HTTP(403, 'not authorized')

        def allow_access(interactivelogin=False):
            row = table(service=session._cas_service, user_id=self.user.id)
            if row:
                ticket = row.ticket
            else:
                ticket = 'ST-' + web2py_uuid()
                table.insert(service=session._cas_service,
                             user_id=self.user.id,
                             ticket=ticket,
                             created_on=request.now,
                             renew=interactivelogin)
            service = session._cas_service
            query_sep = '&' if '?' in service else '?'
            del session._cas_service
            if 'warn' in request.vars and not interactivelogin:
                response.headers[
                    'refresh'] = "5;URL=%s" % service + query_sep + "ticket=" + ticket
                return A("Continue to %s" % service,
                         _href=service + query_sep + "ticket=" + ticket)
            else:
                redirect(service + query_sep + "ticket=" + ticket)
        if self.is_logged_in() and 'renew' not in request.vars:
            return allow_access()
        elif not self.is_logged_in() and 'gateway' in request.vars:
            redirect(session._cas_service)

        def cas_onaccept(form, onaccept=onaccept):
            if onaccept is not DEFAULT:
                onaccept(form)
            return allow_access(interactivelogin=True)
        return self.login(next, onvalidation, cas_onaccept, log)

    def cas_validate(self, version=2, proxy=False):
        request = current.request
        db, table = self.db, self.table_cas()
        current.response.headers['Content-Type'] = 'text'
        ticket = request.vars.ticket
        renew = 'renew' in request.vars
        row = table(ticket=ticket)
        success = False
        if row:
            userfield = self.settings.login_userfield or 'username' \
                if 'username' in table.fields else 'email'
            # If ticket is a service Ticket and RENEW flag respected
            if ticket[0:3] == 'ST-' and \
                    not ((row.renew and renew) ^ renew):
                user = self.table_user()(row.user_id)
                row.delete_record()
                success = True

        def build_response(body):
            xml_body = to_native(TAG['cas:serviceResponse'](
                    body, **{'_xmlns:cas': 'http://www.yale.edu/tp/cas'}).xml())
            return '<?xml version="1.0" encoding="UTF-8"?>\n' + xml_body
        if success:
            if version == 1:
                message = 'yes\n%s' % user[userfield]
            elif version == 3:
                username = user.get('username', user[userfield])
                message = build_response(
                    TAG['cas:authenticationSuccess'](
                        TAG['cas:user'](username),
                        TAG['cas:attributes'](
                            *[TAG['cas:' + field.name](user[field.name])
                              for field in self.table_user()
                              if field.readable])))
            else:  # assume version 2
                username = user.get('username', user[userfield])
                message = build_response(
                    TAG['cas:authenticationSuccess'](
                        TAG['cas:user'](username),
                        *[TAG['cas:' + field.name](user[field.name])
                          for field in self.table_user()
                          if field.readable]))
        else:
            if version == 1:
                message = 'no\n'
            elif row:
                message = build_response(TAG['cas:authenticationFailure']())
            else:
                message = build_response(
                    TAG['cas:authenticationFailure'](
                        'Ticket %s not recognized' % ticket,
                        _code='INVALID TICKET'))
        raise HTTP(200, message)



import smtplib
import logging


class EmailAuthPlugin:

    name = "email_auth"
    label = "Email"

    def __init__(self, server="smtp.gmail.com:587", domain="@gmail.com", tls_mode=None):
        self.server = server
        self.domain = domain
        self.tls_mode = tls_mode

    def validate_credentials(self, username, password):
        email = username
        server = self.server
        domain = self.domain
        tls_mode = self.tls_mode
        if domain:
            if not isinstance(domain, (list, tuple)):
                domain = [str(domain)]
            if not [d for d in domain if email[-len(d) :] == d]:
                return False
        if ":" in server:
            host, port = server.split(":")
        else:
            host = server
            port = 25

        if tls_mode is None:  # then auto detect
            tls_mode = port == "587"
        try:
            server = None
            server = smtplib.SMTP(host, port)
            server.ehlo()
            if tls_mode:
                server.starttls()
                server.ehlo()
            server.login(email, password)
            server.quit()
            return True
        except:
            logging.exception("email_auth() failed")
            if server:
                try:
                    server.quit()
                except:  # server might already close connection after error
                    pass
            return False



from . import OAuth2


class OAuth2Google(OAuth2):
    name = "oauth2google"
    label = "Google"

    login_url = "https://accounts.google.com/o/oauth2/v2/auth"
    token_url = "https://oauth2.googleapis.com/token"
    userinfo_url = "https://www.googleapis.com/oauth2/v3/userinfo"
    revoke_url = "https://accounts.google.com/o/oauth2/revoke"
    default_scope = "email profile"
    maps = {
        "email": "email",
        "sso_id": "sub",
        "first_name": "given_name",
        "last_name": "family_name",
    }



from . import OAuth2


class OAuth2Facebook(OAuth2):
    name = "oauth2facebook"
    label = "Facebook"

    login_url = "https://www.facebook.com/v3.3/dialog/oauth"
    token_url = "https://graph.facebook.com/v3.3/oauth/access_token"
    userinfo_url = "https://graph.facebook.com/me?fields=id,email,first_name,last_name"
    revoke_url = "https://accounts.google.com/me/permissions"
    default_scope = None
    maps = {
        "email": "email",
        "sso_id": "id",
        "first_name": "first_name",
        "last_name": "last_name",
    }

    def revoke(self, token):
        requests.delete(self.revoke_url, headers={"Authorization": token})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: set ts=4 sw=4 et ai:
"""
| This file is part of the web2py Web Framework
| Copyrighted by Massimo Di Pierro <mdipierro@cs.depaul.edu>
| License: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)

Background processes made simple
---------------------------------
"""

from __future__ import print_function

import socket
import os
import logging
import types
from functools import reduce
import datetime
import re
import sys
from json import loads, dumps
import tempfile
import traceback
import threading
import multiprocessing
import time
import signal

from gluon import DAL, Field, IS_NOT_EMPTY, IS_IN_SET, IS_NOT_IN_DB, IS_EMPTY_OR
from gluon import IS_INT_IN_RANGE, IS_DATETIME, IS_IN_DB
from gluon.utils import web2py_uuid
from gluon._compat import (
    Queue,
    long,
    iteritems,
    PY2,
    to_bytes,
    string_types,
    integer_types,
)
from gluon.storage import Storage

USAGE = """
## Example

For any existing application myapp

Create File: myapp/models/scheduler.py ======
from gluon.scheduler import Scheduler

def demo1(*args, **vars):
    print('you passed args=%s and vars=%s' % (args, vars))
    return 'done!'

def demo2():
    1/0

scheduler = Scheduler(db, dict(demo1=demo1, demo2=demo2))
## run worker nodes with:

   cd web2py
   python web2py.py -K myapp
or
   python gluon/scheduler.py -u sqlite://storage.sqlite \
                             -f applications/myapp/databases/ \
                             -t mytasks.py
(-h for info)
python scheduler.py -h

## schedule jobs using
http://127.0.0.1:8000/myapp/appadmin/insert/db/scheduler_task

## monitor scheduled jobs
http://127.0.0.1:8000/myapp/appadmin/select/db?query=db.scheduler_task.id

## view completed jobs
http://127.0.0.1:8000/myapp/appadmin/select/db?query=db.scheduler_run.id

## view workers
http://127.0.0.1:8000/myapp/appadmin/select/db?query=db.scheduler_worker.id

"""

IDENTIFIER = "%s#%s" % (socket.gethostname(), os.getpid())

logger = logging.getLogger("web2py.scheduler.%s" % IDENTIFIER)

QUEUED = "QUEUED"
ASSIGNED = "ASSIGNED"
RUNNING = "RUNNING"
COMPLETED = "COMPLETED"
FAILED = "FAILED"
TIMEOUT = "TIMEOUT"
STOPPED = "STOPPED"
ACTIVE = "ACTIVE"
TERMINATE = "TERMINATE"
DISABLED = "DISABLED"
KILL = "KILL"
PICK = "PICK"
STOP_TASK = "STOP_TASK"
EXPIRED = "EXPIRED"
SECONDS = 1
HEARTBEAT = 3 * SECONDS
MAXHIBERNATION = 10
CLEAROUT = "!clear!"
RESULTINFILE = "result_in_file:"

CALLABLETYPES = (
    types.LambdaType,
    types.FunctionType,
    types.BuiltinFunctionType,
    types.MethodType,
    types.BuiltinMethodType,
)


class Task(object):
    """Defines a "task" object that gets passed from the main thread to the
    executor's one
    """

    def __init__(self, app, function, timeout, args="[]", vars="{}", **kwargs):
        logger.debug(" new task allocated: %s.%s", app, function)
        self.app = app
        self.function = function
        self.timeout = timeout
        self.args = args  # json
        self.vars = vars  # json
        self.__dict__.update(kwargs)

    def __str__(self):
        return "<Task: %s>" % self.function


class TaskReport(object):
    """Defines a "task report" object that gets passed from the executor's
    thread to the main one
    """

    def __init__(self, status, result=None, output=None, tb=None):
        logger.debug("    new task report: %s", status)
        if tb:
            logger.debug("   traceback: %s", tb)
        else:
            logger.debug("   result: %s", result)
        self.status = status
        self.result = result
        self.output = output
        self.tb = tb

    def __str__(self):
        return "<TaskReport: %s>" % self.status


class JobGraph(object):
    """Experimental: dependencies amongs tasks."""

    def __init__(self, db, job_name):
        self.job_name = job_name or "job_0"
        self.db = db

    def add_deps(self, task_parent, task_child):
        """Create a dependency between task_parent and task_child."""
        self.db.scheduler_task_deps.insert(
            task_parent=task_parent, task_child=task_child, job_name=self.job_name
        )

    def validate(self, job_name=None):
        """Validate if all tasks job_name can be completed.

        Checks if there are no mutual dependencies among tasks.
        Commits at the end if successfull, or it rollbacks the entire
        transaction. Handle with care!
        """
        db = self.db
        sd = db.scheduler_task_deps
        if job_name:
            q = sd.job_name == job_name
        else:
            q = sd.id

        edges = db(q).select()
        nested_dict = {}
        for row in edges:
            k = row.task_parent
            if k in nested_dict:
                nested_dict[k].add(row.task_child)
            else:
                nested_dict[k] = set((row.task_child,))
        try:
            rtn = []
            for k, v in nested_dict.items():
                v.discard(k)  # Ignore self dependencies
            extra_items_in_deps = reduce(set.union, nested_dict.values()) - set(
                nested_dict.keys()
            )
            nested_dict.update(dict((item, set()) for item in extra_items_in_deps))
            while True:
                ordered = set(item for item, dep in nested_dict.items() if not dep)
                if not ordered:
                    break
                rtn.append(ordered)
                nested_dict = dict(
                    (item, (dep - ordered))
                    for item, dep in nested_dict.items()
                    if item not in ordered
                )
            assert not nested_dict, (
                "A cyclic dependency exists amongst %r" % nested_dict
            )
            db.commit()
            return rtn
        except Exception:
            db.rollback()
            return None


class CronParser(object):
    def __init__(self, cronline, base=None):
        self.cronline = cronline
        self.sched = base or datetime.datetime.now()
        self.task = None

    @staticmethod
    def _rangetolist(s, period="min"):
        if s.startswith("*"):
            if period == "min":
                s = s.replace("*", "0-59", 1)
            elif period == "hr":
                s = s.replace("*", "0-23", 1)
            elif period == "dom":
                s = s.replace("*", "1-31", 1)
            elif period == "mon":
                s = s.replace("*", "1-12", 1)
            elif period == "dow":
                s = s.replace("*", "0-6", 1)
        match = re.match(r"(\d+)-(\d+)/(\d+)", s)
        if match:
            max_ = int(match.group(2)) + 1
            step_ = int(match.group(3))
        else:
            match = re.match(r"(\d+)/(\d+)", s)
            if match:
                ranges_max = dict(min=59, hr=23, mon=12, dom=31, dow=7)
                max_ = ranges_max[period] + 1
                step_ = int(match.group(2))
        if match:
            min_ = int(match.group(1))
            retval = list(range(min_, max_, step_))
        else:
            retval = []
        return retval

    @staticmethod
    def _sanitycheck(values, period):
        if period == "min":
            check = all(0 <= i <= 59 for i in values)
        elif period == "hr":
            check = all(0 <= i <= 23 for i in values)
        elif period == "dom":
            domrange = list(range(1, 32)) + ["l"]
            check = all(i in domrange for i in values)
        elif period == "mon":
            check = all(1 <= i <= 12 for i in values)
        elif period == "dow":
            check = all(0 <= i <= 7 for i in values)
        return check

    def _parse(self):
        line = self.cronline.lower()
        task = {}
        if line.startswith("@yearly"):
            line = line.replace("@yearly", "0 0 1 1 *")
        elif line.startswith("@annually"):
            line = line.replace("@annually", "0 0 1 1 *")
        elif line.startswith("@monthly"):
            line = line.replace("@monthly", "0 0 1 * *")
        elif line.startswith("@weekly"):
            line = line.replace("@weekly", "0 0 * * 0")
        elif line.startswith("@daily"):
            line = line.replace("@daily", "0 0 * * *")
        elif line.startswith("@midnight"):
            line = line.replace("@midnight", "0 0 * * *")
        elif line.startswith("@hourly"):
            line = line.replace("@hourly", "0 * * * *")
        params = line.strip().split()
        if len(params) < 5:
            raise ValueError("Invalid cron line (too short)")
        elif len(params) > 5:
            raise ValueError("Invalid cron line (too long)")
        daysofweek = {
            "sun": 0,
            "mon": 1,
            "tue": 2,
            "wed": 3,
            "thu": 4,
            "fri": 5,
            "sat": 6,
        }
        monthsofyear = {
            "jan": 1,
            "feb": 2,
            "mar": 3,
            "apr": 4,
            "may": 5,
            "jun": 6,
            "jul": 7,
            "aug": 8,
            "sep": 9,
            "oct": 10,
            "nov": 11,
            "dec": 12,
        }
        for (s, i) in zip(params, ("min", "hr", "dom", "mon", "dow")):
            if s != "*":
                task[i] = []
                vals = s.split(",")
                for val in vals:
                    if i == "dow":
                        refdict = daysofweek
                    elif i == "mon":
                        refdict = monthsofyear
                    if i in ("dow", "mon") and "-" in val and "/" not in val:
                        isnum = val.split("-")[0].isdigit()
                        if isnum:
                            val = "%s/1" % val
                        else:
                            val = "-".join(
                                [str(refdict.get(v, "")) for v in val.split("-")]
                            )
                    if "-" in val and "/" not in val:
                        val = "%s/1" % val
                    if "/" in val:
                        task[i] += self._rangetolist(val, i)
                    elif val.isdigit():
                        task[i].append(int(val))
                    elif i in ("dow", "mon"):
                        if val in refdict:
                            task[i].append(refdict[val])
                    elif i == "dom" and val == "l":
                        task[i].append(val)
                if not task[i]:
                    raise ValueError("Invalid cron value (%s)" % s)
                if not self._sanitycheck(task[i], i):
                    raise ValueError("Invalid cron value (%s)" % s)
                task[i] = sorted(task[i])
        self.task = task

    @staticmethod
    def _get_next_dow(sched, task):
        task_dow = [a % 7 for a in task["dow"]]
        while sched.isoweekday() % 7 not in task_dow:
            sched += datetime.timedelta(days=1)
        return sched

    @staticmethod
    def _get_next_dom(sched, task):
        if task["dom"] == ["l"]:
            # instead of calendar.isleap
            try:
                last_feb = 29
                datetime.date(sched.year, 2, last_feb)
            except ValueError:
                last_feb = 28
            lastdayofmonth = [31, last_feb, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
            task_dom = [lastdayofmonth[sched.month - 1]]
        else:
            task_dom = task["dom"]
        while sched.day not in task_dom:
            sched += datetime.timedelta(days=1)
        return sched

    @staticmethod
    def _get_next_mon(sched, task):
        while sched.month not in task["mon"]:
            if sched.month < 12:
                sched = sched.replace(month=sched.month + 1)
            else:
                sched = sched.replace(month=1, year=sched.year + 1)
        return sched

    @staticmethod
    def _getnext_hhmm(sched, task, add_to=True):
        if add_to:
            sched += datetime.timedelta(minutes=1)
        if "min" in task:
            while sched.minute not in task["min"]:
                sched += datetime.timedelta(minutes=1)
        if "hr" in task and sched.hour not in task["hr"]:
            while sched.hour not in task["hr"]:
                sched += datetime.timedelta(hours=1)
        return sched

    def _getnext_date(self, sched, task):
        if "dow" in task and "dom" in task:
            dow = self._get_next_dow(sched, task)
            dom = self._get_next_dom(sched, task)
            sched = min(dow, dom)
        elif "dow" in task:
            sched = self._get_next_dow(sched, task)
        elif "dom" in task:
            sched = self._get_next_dom(sched, task)
        if "mon" in task:
            sched = self._get_next_mon(sched, task)
        return sched.replace(hour=0, minute=0)

    def next(self):
        """Get next date according to specs."""
        if not self.task:
            self._parse()
        task = self.task
        sched = self.sched
        x = 0
        while x < 1000:  # avoid potential max recursions
            x += 1
            try:
                next_date = self._getnext_date(sched, task)
            except (ValueError, OverflowError) as e:
                raise ValueError("Invalid cron expression (%s)" % e)
            if next_date.date() > self.sched.date():
                # we rolled date, check for valid hhmm
                sched = self._getnext_hhmm(next_date, task, False)
                break
            else:
                # same date, get next hhmm
                sched_time = self._getnext_hhmm(sched, task, True)
                if sched_time.date() > sched.date():
                    # we rolled date again :(
                    sched = sched_time
                else:
                    sched = sched_time
                    break
        else:
            raise ValueError(
                "Potential bug found, please submit your "
                "cron expression to the authors"
            )
        self.sched = sched
        return sched

    def __iter__(self):
        """Support iteration."""
        return self

    __next__ = next


# the two functions below deal with simplejson decoding as unicode,
# esp for the dict decode and subsequent usage as function Keyword arguments
# unicode variable names won't work!
# borrowed from http://stackoverflow.com/questions/956867/


def _decode_list(lst):
    if not PY2:
        return lst
    newlist = []
    for i in lst:
        if isinstance(i, string_types):
            i = to_bytes(i)
        elif isinstance(i, list):
            i = _decode_list(i)
        newlist.append(i)
    return newlist


def _decode_dict(dct):
    if not PY2:
        return dct
    newdict = {}
    for k, v in iteritems(dct):
        k = to_bytes(k)
        if isinstance(v, string_types):
            v = to_bytes(v)
        elif isinstance(v, list):
            v = _decode_list(v)
        newdict[k] = v
    return newdict


def executor(retq, task, outq):
    """The function used to execute tasks in the background process."""
    logger.debug("    task started")

    class LogOutput(object):
        """Facility to log output at intervals."""

        def __init__(self, out_queue):
            self.out_queue = out_queue
            self.stdout = sys.stdout
            self.written = False
            sys.stdout = self

        def close(self):
            sys.stdout = self.stdout
            if self.written:
                # see "Joining processes that use queues" section in
                # https://docs.python.org/2/library/multiprocessing.html#programming-guidelines
                # https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
                self.out_queue.cancel_join_thread()

        def flush(self):
            pass

        def write(self, data):
            self.out_queue.put(data)
            self.written = True

    W2P_TASK = Storage({"id": task.task_id, "uuid": task.uuid, "run_id": task.run_id})
    stdout = LogOutput(outq)
    try:
        if task.app:
            from gluon.shell import env, parse_path_info
            from gluon import current

            ## FIXME: why temporarily change the log level of the root logger?
            # level = logging.getLogger().getEffectiveLevel()
            # logging.getLogger().setLevel(logging.WARN)
            # support for task.app like 'app/controller'
            (a, c, f) = parse_path_info(task.app)
            _env = env(
                a=a, c=c, import_models=True, extra_request={"is_scheduler": True}
            )
            # logging.getLogger().setLevel(level)
            f = task.function
            functions = current._scheduler.tasks
            if functions:
                _function = functions.get(f)
            else:
                # look into env
                _function = _env.get(f)
            if not isinstance(_function, CALLABLETYPES):
                raise NameError("name '%s' not found in scheduler's environment" % f)
            # Inject W2P_TASK into environment
            _env.update({"W2P_TASK": W2P_TASK})
            # Inject W2P_TASK into current
            current.W2P_TASK = W2P_TASK
            globals().update(_env)
            args = _decode_list(loads(task.args))
            vars = loads(task.vars, object_hook=_decode_dict)
            result = dumps(_function(*args, **vars))
        else:
            # for testing purpose only
            result = eval(task.function)(
                *loads(task.args, object_hook=_decode_dict),
                **loads(task.vars, object_hook=_decode_dict)
            )
        if len(result) >= 1024:
            fd, temp_path = tempfile.mkstemp(suffix=".w2p_sched")
            with os.fdopen(fd, "w") as f:
                f.write(result)
            result = RESULTINFILE + temp_path
        retq.put(TaskReport("COMPLETED", result=result))
    except:
        tb = traceback.format_exc()
        retq.put(TaskReport("FAILED", tb=tb))
    finally:
        stdout.close()


class IS_CRONLINE(object):
    """
    Validates cronline
    """

    def __init__(self, error_message=None):
        self.error_message = error_message

    def __call__(self, value, record_id=None):
        recur = CronParser(value, datetime.datetime.now())
        try:
            recur.next()
            return (value, None)
        except ValueError as e:
            if not self.error_message:
                return (value, e)
            return (value, self.error_message)


class TYPE(object):
    """
    Validator that checks whether field is valid json and validates its type.
    Used for `args` and `vars` of the scheduler_task table
    """

    def __init__(self, myclass=list, parse=False):
        self.myclass = myclass
        self.parse = parse

    def __call__(self, value, record_id=None):
        from gluon import current

        try:
            obj = loads(value)
        except:
            return (value, current.T("invalid json"))
        else:
            if isinstance(obj, self.myclass):
                if self.parse:
                    return (obj, None)
                else:
                    return (value, None)
            else:
                return (value, current.T("Not of type: %s") % self.myclass)


TASK_STATUS = (QUEUED, RUNNING, COMPLETED, FAILED, TIMEOUT, STOPPED, EXPIRED)
RUN_STATUS = (RUNNING, COMPLETED, FAILED, TIMEOUT, STOPPED)
WORKER_STATUS = (ACTIVE, PICK, DISABLED, TERMINATE, KILL, STOP_TASK)


class Scheduler(threading.Thread):
    """Scheduler object

    Args:
        db: DAL connection where Scheduler will create its tables
        tasks(dict): either a dict containing name-->func or None.
            If None, functions will be searched in the environment
        migrate(bool): turn migration on/off for the Scheduler's tables
        worker_name(str): force worker_name to identify each process.
            Leave it to None to autoassign a name (hostname#pid)
        group_names(list): process tasks belonging to this group
            defaults to ['main'] if nothing gets passed
        heartbeat(int): how many seconds the worker sleeps between one
            execution and the following one. Indirectly sets how many seconds
            will pass between checks for new tasks
        max_empty_runs(int): how many loops are allowed to pass without
            processing any tasks before exiting the process. 0 to keep always
            the process alive
        discard_results(bool): Scheduler stores executions's details into the
            scheduler_run table. By default, only if there is a result the
            details are kept. Turning this to True means discarding results
            even for tasks that return something
        utc_time(bool): do all datetime calculations assuming UTC as the
            timezone. Remember to pass `start_time` and `stop_time` to tasks
            accordingly
        use_spawn(bool): use spawn for subprocess (only useable with python3)
    """

    def __init__(
        self,
        db,
        tasks=None,
        migrate=True,
        worker_name=None,
        group_names=None,
        heartbeat=HEARTBEAT,
        max_empty_runs=0,
        discard_results=False,
        utc_time=False,
        use_spawn=False,
    ):

        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.process = None  # the background process
        self.process_queues = (None, None)
        self.have_heartbeat = True  # set to False to kill
        self.empty_runs = 0

        self.db = db
        self.db_thread = None
        self.tasks = tasks
        self.group_names = group_names or ["main"]
        self.heartbeat = heartbeat
        self.worker_name = worker_name or IDENTIFIER
        self.max_empty_runs = max_empty_runs
        self.discard_results = discard_results
        self.is_a_ticker = False
        self.do_assign_tasks = False
        self.greedy = False
        self.utc_time = utc_time
        self.w_stats_lock = threading.RLock()
        self.w_stats = Storage(
            dict(
                status=RUNNING,
                sleep=heartbeat,
                total=0,
                errors=0,
                empty_runs=0,
                queue=0,
                distribution=None,
                workers=0,
            )
        )  # dict holding statistics

        from gluon import current

        current._scheduler = self

        self.define_tables(db, migrate=migrate)
        self.use_spawn = use_spawn

    def execute(self, task):
        """Start the background process.

        Args:
            task : a `Task` object

        Returns:
            a `TaskReport` object
        """
        outq = None
        retq = None
        if self.use_spawn and not PY2:
            ctx = multiprocessing.get_context("spawn")
            outq = ctx.Queue()
            retq = ctx.Queue(maxsize=1)
            self.process = p = ctx.Process(target=executor, args=(retq, task, outq))
        else:
            outq = multiprocessing.Queue()
            retq = multiprocessing.Queue(maxsize=1)
            self.process = p = multiprocessing.Process(
                target=executor, args=(retq, task, outq)
            )

        self.process_queues = (retq, outq)

        logger.debug("   task starting")
        p.start()
        start = time.time()

        if task.sync_output > 0:
            run_timeout = task.sync_output
        else:
            run_timeout = task.timeout
        task_output = tout = ""
        try:
            while p.is_alive() and (
                not task.timeout or time.time() - start < task.timeout
            ):
                # NOTE: try always to empty the out queue before
                #       the child process is joined,
                #       see "Joining processes that use queues" section in
                # https://docs.python.org/2/library/multiprocessing.html#programming-guidelines
                # https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
                while True:
                    try:
                        tout += outq.get(timeout=2)
                    except Queue.Empty:
                        break
                if tout:
                    logger.debug(' partial output: "%s"', tout)
                    if CLEAROUT in tout:
                        task_output = tout[tout.rfind(CLEAROUT) + len(CLEAROUT) :]
                    else:
                        task_output += tout
                    try:
                        db = self.db
                        db(db.scheduler_run.id == task.run_id).update(
                            run_output=task_output
                        )
                        db.commit()
                        tout = ""
                        logger.debug(" partial output saved")
                    except Exception:
                        logger.exception(" error while saving partial output")
                        task_output = task_output[: -len(tout)]
                p.join(timeout=run_timeout)
        except:
            logger.exception("    task stopped by general exception")
            self.terminate_process()
            tr = TaskReport(STOPPED)
        else:
            if p.is_alive():
                logger.debug("    task timeout")
                self.terminate_process(flush_ret=False)
                try:
                    # we try to get a traceback here
                    tr = retq.get(timeout=2)  # NOTE: risky after terminate
                    tr.status = TIMEOUT
                    tr.output = task_output
                except Queue.Empty:
                    tr = TaskReport(TIMEOUT)
            else:
                try:
                    tr = retq.get_nowait()
                except Queue.Empty:
                    logger.debug("    task stopped")
                    tr = TaskReport(STOPPED)
                else:
                    logger.debug("  task completed or failed")
        result = tr.result
        if result and result.startswith(RESULTINFILE):
            temp_path = result.replace(RESULTINFILE, "", 1)
            with open(temp_path) as f:
                tr.result = f.read()
            os.unlink(temp_path)
        tr.output = task_output
        return tr

    _terminate_process_lock = threading.RLock()

    def terminate_process(self, flush_out=True, flush_ret=True):
        """Terminate any running tasks (internal use only)"""
        if self.process is not None:
            # must synchronize since we are called by main and heartbeat thread
            with self._terminate_process_lock:
                if flush_out:
                    queue = self.process_queues[1]
                    while not queue.empty():  # NOTE: empty() is not reliable
                        try:
                            queue.get_nowait()
                        except Queue.Empty:
                            pass
                if flush_ret:
                    queue = self.process_queues[0]
                    while not queue.empty():
                        try:
                            queue.get_nowait()
                        except Queue.Empty:
                            pass
                logger.debug("terminating process")
                try:
                    # NOTE: terminate should not be called when using shared
                    #       resources, see "Avoid terminating processes"
                    #       section in
                    # https://docs.python.org/2/library/multiprocessing.html#programming-guidelines
                    # https://docs.python.org/3/library/multiprocessing.html#programming-guidelines
                    self.process.terminate()
                    # NOTE: calling join after a terminate is risky,
                    #       as explained in "Avoid terminating processes"
                    #       section this can lead to a deadlock
                    self.process.join()
                finally:
                    self.process = None

    def die(self):
        """Forces termination of the worker process along with any running
        task"""
        logger.info("die!")
        self.have_heartbeat = False
        self.terminate_process()

    def give_up(self):
        """Waits for any running task to be executed, then exits the worker
        process"""
        logger.info("Giving up as soon as possible!")
        self.have_heartbeat = False

    def run(self):
        """This is executed by the heartbeat thread"""
        counter = 0
        while self.have_heartbeat:
            self.send_heartbeat(counter)
            counter += 1

    def start_heartbeats(self):
        self.start()

    def __get_migrate(self, tablename, migrate=True):
        if migrate is False:
            return False
        elif migrate is True:
            return True
        elif isinstance(migrate, str):
            return "%s%s.table" % (migrate, tablename)
        return True

    def now(self):
        """Shortcut that fetches current time based on UTC preferences."""
        return self.utc_time and datetime.datetime.utcnow() or datetime.datetime.now()

    def set_requirements(self, scheduler_task):
        """Called to set defaults for lazy_tables connections."""
        from gluon import current

        if hasattr(current, "request"):
            scheduler_task.application_name.default = "%s/%s" % (
                current.request.application,
                current.request.controller,
            )

    def define_tables(self, db, migrate):
        """Define Scheduler tables structure."""
        from pydal.base import DEFAULT

        logger.debug("defining tables (migrate=%s)", migrate)
        now = self.now
        db.define_table(
            "scheduler_task",
            Field(
                "application_name",
                requires=IS_NOT_EMPTY(),
                default=None,
                writable=False,
            ),
            Field("task_name", default=None),
            Field("group_name", default="main"),
            Field(
                "status",
                requires=IS_IN_SET(TASK_STATUS),
                default=QUEUED,
                writable=False,
            ),
            Field("broadcast", "boolean", default=False),
            Field(
                "function_name",
                requires=IS_IN_SET(sorted(self.tasks.keys()))
                if self.tasks
                else DEFAULT,
            ),
            Field(
                "uuid",
                length=255,
                requires=IS_NOT_IN_DB(db, "scheduler_task.uuid"),
                unique=True,
                default=web2py_uuid,
            ),
            Field("args", "text", default="[]", requires=TYPE(list)),
            Field("vars", "text", default="{}", requires=TYPE(dict)),
            Field("enabled", "boolean", default=True),
            Field("start_time", "datetime", default=now, requires=IS_DATETIME()),
            Field("next_run_time", "datetime", default=now),
            Field("stop_time", "datetime"),
            Field(
                "repeats",
                "integer",
                default=1,
                comment="0=unlimited",
                requires=IS_INT_IN_RANGE(0, None),
            ),
            Field(
                "retry_failed",
                "integer",
                default=0,
                comment="-1=unlimited",
                requires=IS_INT_IN_RANGE(-1, None),
            ),
            Field(
                "period",
                "integer",
                default=60,
                comment="seconds",
                requires=IS_INT_IN_RANGE(0, None),
            ),
            Field(
                "prevent_drift",
                "boolean",
                default=False,
                comment="Exact start_times between runs",
            ),
            Field(
                "cronline",
                default=None,
                comment='Discard "period", use this cron expr instead',
                requires=IS_EMPTY_OR(IS_CRONLINE()),
            ),
            Field(
                "timeout",
                "integer",
                default=60,
                comment="seconds",
                requires=IS_INT_IN_RANGE(1, None),
            ),
            Field(
                "sync_output",
                "integer",
                default=0,
                comment="update output every n sec: 0=never",
                requires=IS_INT_IN_RANGE(0, None),
            ),
            Field("times_run", "integer", default=0, writable=False),
            Field("times_failed", "integer", default=0, writable=False),
            Field("last_run_time", "datetime", writable=False, readable=False),
            Field("assigned_worker_name", default="", writable=False),
            on_define=self.set_requirements,
            migrate=self.__get_migrate("scheduler_task", migrate),
            format="(%(id)s) %(task_name)s",
        )

        db.define_table(
            "scheduler_run",
            Field("task_id", "reference scheduler_task"),
            Field("status", requires=IS_IN_SET(RUN_STATUS)),
            Field("start_time", "datetime"),
            Field("stop_time", "datetime"),
            Field("run_output", "text"),
            Field("run_result", "text"),
            Field("traceback", "text"),
            Field("worker_name", default=self.worker_name),
            migrate=self.__get_migrate("scheduler_run", migrate),
        )

        db.define_table(
            "scheduler_worker",
            Field("worker_name", length=255, unique=True),
            Field("first_heartbeat", "datetime"),
            Field("last_heartbeat", "datetime"),
            Field("status", requires=IS_IN_SET(WORKER_STATUS)),
            Field("is_ticker", "boolean", default=False, writable=False),
            Field("group_names", "list:string", default=self.group_names),
            Field("worker_stats", "json"),
            migrate=self.__get_migrate("scheduler_worker", migrate),
        )

        db.define_table(
            "scheduler_task_deps",
            Field("job_name", default="job_0"),
            Field(
                "task_parent",
                "integer",
                requires=IS_IN_DB(db, "scheduler_task.id", "%(task_name)s"),
            ),
            Field("task_child", "reference scheduler_task"),
            Field("can_visit", "boolean", default=False),
            migrate=self.__get_migrate("scheduler_task_deps", migrate),
        )

        if migrate is not False:
            db.commit()

    def loop(self, worker_name=None):
        """Main loop.

        This works basically as a neverending loop that:

        - checks if the worker is ready to process tasks (is not DISABLED)
        - pops a task from the queue
        - if there is a task:

          - spawns the executor background process
          - waits for the process to be finished
          - sleeps `heartbeat` seconds
        - if there is not a task:

          - checks for max_empty_runs
          - sleeps `heartbeat` seconds

        """
        signal.signal(signal.SIGTERM, lambda signum, stack_frame: sys.exit(1))
        try:
            self.start_heartbeats()
            while self.have_heartbeat:
                with self.w_stats_lock:
                    is_disabled = self.w_stats.status == DISABLED
                if is_disabled:
                    logger.debug(
                        "Someone stopped me, sleeping until better" " times come (%s)",
                        self.w_stats.sleep,
                    )
                    self.sleep()
                    continue
                logger.debug("looping...")
                if self.is_a_ticker and self.do_assign_tasks:
                    # I'm a ticker, and 5 loops passed without
                    # reassigning tasks, let's do that
                    self.wrapped_assign_tasks()
                task = self.wrapped_pop_task()
                if task:
                    with self.w_stats_lock:
                        self.w_stats.empty_runs = 0
                        self.w_stats.status = RUNNING
                        self.w_stats.total += 1
                    self.wrapped_report_task(task, self.execute(task))
                    with self.w_stats_lock:
                        if not self.w_stats.status == DISABLED:
                            self.w_stats.status = ACTIVE
                else:
                    with self.w_stats_lock:
                        self.w_stats.empty_runs += 1
                        if self.max_empty_runs != 0:
                            logger.debug(
                                "empty runs %s/%s",
                                self.w_stats.empty_runs,
                                self.max_empty_runs,
                            )
                            if self.w_stats.empty_runs >= self.max_empty_runs:
                                logger.info("empty runs limit reached, killing myself")
                                self.die()
                    if self.is_a_ticker and self.greedy:
                        # there could be other tasks ready to be assigned
                        logger.info("TICKER: greedy loop")
                        self.wrapped_assign_tasks()
                    logger.debug("sleeping...")
                    self.sleep()
        except (KeyboardInterrupt, SystemExit):
            logger.info("catched")
            self.die()

    def wrapped_pop_task(self):
        """Commodity function to call `pop_task` and trap exceptions.

        If an exception is raised, assume it happened because of database
        contention and retries `pop_task` after 0.5 seconds
        """
        db = self.db
        db.commit()  # for MySQL only; FIXME: Niphlod, still needed? could avoid when not MySQL?
        for x in range(10):
            try:
                return self.pop_task()
            except Exception:
                logger.exception("    error popping tasks")
                self.w_stats.errors += 1
                db.rollback()
                time.sleep(0.5)

    def pop_task(self):
        """Grab a task ready to be executed from the queue."""
        now = self.now()
        db = self.db
        st = db.scheduler_task
        grabbed = db(
            (st.assigned_worker_name == self.worker_name) & (st.status == ASSIGNED)
        )

        task = grabbed.select(limitby=(0, 1), orderby=st.next_run_time).first()
        if task:
            # none will touch my task!
            task.update_record(status=RUNNING, last_run_time=now)
            db.commit()
            logger.debug("   work to do %s", task.id)
        else:
            logger.info("nothing to do")
            return None

        if task.cronline:
            cron_recur = CronParser(task.cronline, now.replace(second=0, microsecond=0))
            next_run_time = cron_recur.next()
        elif not task.prevent_drift:
            next_run_time = task.last_run_time + datetime.timedelta(seconds=task.period)
        else:
            # calc next_run_time based on available slots
            # see #1191
            next_run_time = task.start_time
            secondspassed = (now - next_run_time).total_seconds()
            times = secondspassed // task.period + 1
            next_run_time += datetime.timedelta(seconds=task.period * times)

        times_run = task.times_run + 1
        if times_run < task.repeats or task.repeats == 0:
            # need to run (repeating task)
            run_again = True
        else:
            # no need to run again
            run_again = False
        run_id = 0
        while not self.discard_results:  # FIXME: forever?
            logger.debug("    new scheduler_run record")
            try:
                run_id = db.scheduler_run.insert(
                    task_id=task.id,
                    status=RUNNING,
                    start_time=now,
                    worker_name=self.worker_name,
                )
                db.commit()
                break
            except Exception:
                logger.exception("    error inserting scheduler_run")
                db.rollback()
                time.sleep(0.5)
        logger.info(
            'new task %(id)s "%(task_name)s"'
            " %(application_name)s.%(function_name)s" % task
        )
        return Task(
            app=task.application_name,
            function=task.function_name,
            timeout=task.timeout,
            args=task.args,  # in json
            vars=task.vars,  # in json
            task_id=task.id,
            run_id=run_id,
            run_again=run_again,
            next_run_time=next_run_time,
            times_run=times_run,
            stop_time=task.stop_time,
            retry_failed=task.retry_failed,
            times_failed=task.times_failed,
            sync_output=task.sync_output,
            uuid=task.uuid,
        )

    def wrapped_report_task(self, task, task_report):
        """Commodity function to call `report_task` and trap exceptions.

        If an exception is raised, assume it happened because of database
        contention and retries `pop_task` after 0.5 seconds
        """
        db = self.db
        while True:  # FIXME: forever?
            try:
                self.report_task(task, task_report)
                db.commit()
                break
            except Exception:
                logger.exception("    error storing result")
                db.rollback()
                time.sleep(0.5)

    def report_task(self, task, task_report):
        """Take care of storing the result according to preferences.

        Deals with logic for repeating tasks.
        """
        now = self.now()
        db = self.db
        st = db.scheduler_task
        sr = db.scheduler_run
        if not self.discard_results:
            if task_report.result != "null" or task_report.tb:
                # result is 'null' as a string if task completed
                # if it's stopped it's None as NoneType, so we record
                # the STOPPED "run" anyway
                logger.debug(" recording task report in db (%s)", task_report.status)
                db(sr.id == task.run_id).update(
                    status=task_report.status,
                    stop_time=now,
                    run_result=task_report.result,
                    run_output=task_report.output,
                    traceback=task_report.tb,
                )
            else:
                logger.debug(" deleting task report in db because of no result")
                db(sr.id == task.run_id).delete()
        # if there is a stop_time and the following run would exceed it
        is_expired = task.stop_time and task.next_run_time > task.stop_time or False
        status = (
            task.run_again
            and is_expired
            and EXPIRED
            or task.run_again
            and not is_expired
            and QUEUED
            or COMPLETED
        )
        if task_report.status == COMPLETED:
            d = dict(
                status=status,
                next_run_time=task.next_run_time,
                times_run=task.times_run,
                times_failed=0,
            )
            db(st.id == task.task_id).update(**d)
            if status == COMPLETED:
                self.update_dependencies(task.task_id)
        else:
            st_mapping = {
                "FAILED": "FAILED",
                "TIMEOUT": "TIMEOUT",
                "STOPPED": "FAILED",
            }[task_report.status]
            status = (
                task.retry_failed
                and task.times_failed < task.retry_failed
                and QUEUED
                or task.retry_failed == -1
                and QUEUED
                or st_mapping
            )
            db(st.id == task.task_id).update(
                times_failed=st.times_failed + 1,
                next_run_time=task.next_run_time,
                status=status,
            )
        logger.info("task completed (%s)", task_report.status)

    def update_dependencies(self, task_id):
        """Unblock execution paths for Jobs."""
        db = self.db
        db(db.scheduler_task_deps.task_child == task_id).update(can_visit=True)

    def adj_hibernation(self):
        """Used to increase the "sleep" interval for DISABLED workers."""
        with self.w_stats_lock:
            if self.w_stats.status == DISABLED:
                wk_st = self.w_stats.sleep
                hibernation = (
                    wk_st + HEARTBEAT if wk_st < MAXHIBERNATION else MAXHIBERNATION
                )
                self.w_stats.sleep = hibernation

    def send_heartbeat(self, counter):
        """Coordination among available workers.

        It:
        - sends the heartbeat
        - elects a ticker among available workers (the only process that
            effectively dispatch tasks to workers)
        - deals with worker's statuses
        - does "housecleaning" for dead workers
        - triggers tasks assignment to workers
        """
        if self.db_thread:
            # BKR 20180612 check if connection still works
            try:
                self.db_thread(self.db_thread.scheduler_worker).count()
            except self.db_thread._adapter.connection.OperationalError:
                # if not -> throw away self.db_thread and force reconnect
                self.db_thread = None

        if not self.db_thread:
            logger.debug("thread building own DAL object")
            self.db_thread = DAL(
                self.db._uri, folder=self.db._adapter.folder, decode_credentials=True
            )
            self.define_tables(self.db_thread, migrate=False)

        try:
            now = self.now()
            db = self.db_thread
            sw = db.scheduler_worker
            st = db.scheduler_task
            # record heartbeat
            row = db(sw.worker_name == self.worker_name).select().first()
            with self.w_stats_lock:
                if not row:
                    sw.insert(
                        status=ACTIVE,
                        worker_name=self.worker_name,
                        first_heartbeat=now,
                        last_heartbeat=now,
                        group_names=self.group_names,
                        worker_stats=self.w_stats,
                    )
                    self.w_stats.status = ACTIVE
                    self.w_stats.sleep = self.heartbeat
                    backed_status = ACTIVE
                else:
                    backed_status = row.status
                    if backed_status == DISABLED:
                        # keep sleeping
                        self.w_stats.status = DISABLED
                        logger.debug("........recording heartbeat (DISABLED)")
                        db(sw.worker_name == self.worker_name).update(
                            last_heartbeat=now, worker_stats=self.w_stats
                        )
                    elif backed_status == TERMINATE:
                        self.w_stats.status = TERMINATE
                        logger.debug("Waiting to terminate the current task")
                        self.give_up()
                    elif backed_status == KILL:
                        self.w_stats.status = KILL
                        self.die()
                        return
                    else:
                        if backed_status == STOP_TASK:
                            logger.info("Asked to kill the current task")
                            self.terminate_process()
                        logger.debug(
                            "........recording heartbeat (%s)", self.w_stats.status
                        )
                        db(sw.worker_name == self.worker_name).update(
                            last_heartbeat=now, status=ACTIVE, worker_stats=self.w_stats
                        )
                        self.w_stats.sleep = self.heartbeat  # re-activating the process
                        if self.w_stats.status != RUNNING:
                            self.w_stats.status = ACTIVE

            self.do_assign_tasks = False
            if counter % 5 == 0 or backed_status == PICK:
                try:
                    # delete dead workers
                    expiration = now - datetime.timedelta(seconds=self.heartbeat * 3)
                    departure = now - datetime.timedelta(
                        seconds=self.heartbeat * 3 * 15
                    )
                    logger.debug("    freeing workers that have not sent heartbeat")
                    dead_workers = db(
                        ((sw.last_heartbeat < expiration) & (sw.status == ACTIVE))
                        | ((sw.last_heartbeat < departure) & (sw.status != ACTIVE))
                    )
                    dead_workers_name = dead_workers._select(sw.worker_name)
                    db(
                        (st.assigned_worker_name.belongs(dead_workers_name))
                        & (st.status == RUNNING)
                    ).update(assigned_worker_name="", status=QUEUED)
                    dead_workers.delete()
                    try:
                        self.is_a_ticker = self.being_a_ticker()
                    except:
                        logger.exception("Error coordinating TICKER")
                    with self.w_stats_lock:
                        if self.w_stats.status == ACTIVE:
                            self.do_assign_tasks = True
                except:
                    logger.exception("Error cleaning up")

            db.commit()
        except:
            logger.exception("Error retrieving status")
            db.rollback()
        self.adj_hibernation()
        self.sleep()

    def being_a_ticker(self):
        """Elect a TICKER process that assigns tasks to available workers.

        Does its best to elect a worker that is not busy processing other tasks
        to allow a proper distribution of tasks among all active workers ASAP
        """
        db = self.db_thread
        sw = db.scheduler_worker
        my_name = self.worker_name
        all_active = db((sw.worker_name != my_name) & (sw.status == ACTIVE)).select(
            sw.is_ticker, sw.worker_name
        )
        ticker = all_active.find(lambda row: row.is_ticker is True).first()
        with self.w_stats_lock:
            not_busy = self.w_stats.status == ACTIVE
        if not ticker:
            # if no other tickers are around
            if not_busy:
                # only if I'm not busy
                db(sw.worker_name == my_name).update(is_ticker=True)
                db(sw.worker_name != my_name).update(is_ticker=False)
                logger.info("TICKER: I'm a ticker")
            else:
                # I'm busy
                if len(all_active) >= 1:
                    # so I'll "downgrade" myself to a "poor worker"
                    db(sw.worker_name == my_name).update(is_ticker=False)
                else:
                    not_busy = True
            db.commit()
            return not_busy
        else:
            logger.info("%s is a ticker, I'm a poor worker" % ticker.worker_name)
            return False

    def wrapped_assign_tasks(self):
        """Commodity function to call `assign_tasks` and trap exceptions.

        If an exception is raised, assume it happened because of database
        contention and retries `assign_task` after 0.5 seconds
        """
        logger.debug("Assigning tasks...")
        db = self.db
        db.commit()  # for MySQL only; FIXME: Niphlod, still needed? could avoid when not MySQL?
        for x in range(10):
            try:
                self.assign_tasks()
                db.commit()
                logger.debug("Tasks assigned...")
                break
            except Exception:
                logger.exception("TICKER: error assigning tasks")
                self.w_stats.errors += 1
                db.rollback()
                time.sleep(0.5)

    def assign_tasks(self):
        """Assign task to workers, that can then pop them from the queue.

        Deals with group_name(s) logic, in order to assign linearly tasks
        to available workers for those groups
        """
        now = self.now()
        db = self.db
        sw = db.scheduler_worker
        st = db.scheduler_task
        sd = db.scheduler_task_deps
        all_workers = db(sw.status == ACTIVE).select()
        # build workers as dict of groups
        wkgroups = {}
        for w in all_workers:
            if w.worker_stats["status"] == "RUNNING":
                continue
            group_names = w.group_names
            for gname in group_names:
                if gname not in wkgroups:
                    wkgroups[gname] = dict(workers=[{"name": w.worker_name, "c": 0}])
                else:
                    wkgroups[gname]["workers"].append({"name": w.worker_name, "c": 0})
        # set queued tasks that expired between "runs" (i.e., you turned off
        # the scheduler): then it wasn't expired, but now it is
        db((st.status.belongs((QUEUED, ASSIGNED))) & (st.stop_time < now)).update(
            status=EXPIRED
        )

        # calculate dependencies
        deps_with_no_deps = db(
            (sd.can_visit == False)
            & (
                ~sd.task_child.belongs(
                    db(sd.can_visit == False)._select(sd.task_parent)
                )
            )
        )._select(sd.task_child)
        no_deps = db(
            (st.status.belongs((QUEUED, ASSIGNED)))
            & ((sd.id == None) | (st.id.belongs(deps_with_no_deps)))
        )._select(
            st.id,
            distinct=True,
            left=sd.on((st.id == sd.task_parent) & (sd.can_visit == False)),
        )

        all_available = db(
            (st.status.belongs((QUEUED, ASSIGNED)))
            & (st.next_run_time <= now)
            & (st.enabled == True)
            & (st.id.belongs(no_deps))
        )

        limit = len(all_workers) * (50 / (len(wkgroups) or 1))
        # if there are a moltitude of tasks, let's figure out a maximum of
        # tasks per worker. This can be further tuned with some added
        # intelligence (like esteeming how many tasks will a worker complete
        # before the ticker reassign them around, but the gain is quite small
        # 50 is a sweet spot also for fast tasks, with sane heartbeat values
        # NB: ticker reassign tasks every 5 cycles, so if a worker completes
        # its 50 tasks in less than heartbeat*5 seconds,
        # it won't pick new tasks until heartbeat*5 seconds pass.

        # If a worker is currently elaborating a long task, its tasks needs to
        # be reassigned to other workers
        # this shuffles up things a bit, in order to give a task equal chances
        # to be executed

        # let's freeze it up
        db.commit()
        tnum = 0
        for group in wkgroups.keys():
            tasks = all_available(st.group_name == group).select(
                limitby=(0, limit), orderby=st.next_run_time
            )
            # let's break up the queue evenly among workers
            for task in tasks:
                tnum += 1
                gname = task.group_name
                ws = wkgroups.get(gname)
                if ws:
                    if task.broadcast:
                        for worker in ws["workers"]:
                            new_task = db.scheduler_task.insert(
                                application_name=task.application_name,
                                task_name=task.task_name,
                                group_name=task.group_name,
                                status=ASSIGNED,
                                broadcast=False,
                                function_name=task.function_name,
                                args=task.args,
                                start_time=now,
                                repeats=1,
                                retry_failed=task.retry_failed,
                                sync_output=task.sync_output,
                                assigned_worker_name=worker["name"],
                            )
                        if task.period:
                            next_run_time = now + datetime.timedelta(
                                seconds=task.period
                            )
                        else:
                            # must be cronline
                            cron_recur = CronParser(
                                task.cronline, now.replace(second=0, microsecond=0)
                            )
                            next_run_time = cron_recur.next()
                        db(st.id == task.id).update(
                            times_run=task.times_run + 1,
                            next_run_time=next_run_time,
                            last_run_time=now,
                        )
                        db.commit()
                    else:
                        counter = 0
                        myw = 0
                        for i, w in enumerate(ws["workers"]):
                            if w["c"] < counter:
                                myw = i
                            counter = w["c"]
                        assigned_wn = wkgroups[gname]["workers"][myw]["name"]
                        d = dict(status=ASSIGNED, assigned_worker_name=assigned_wn)
                        db(
                            (st.id == task.id) & (st.status.belongs((QUEUED, ASSIGNED)))
                        ).update(**d)
                        wkgroups[gname]["workers"][myw]["c"] += 1
                db.commit()
        # I didn't report tasks but I'm working nonetheless!!!!
        with self.w_stats_lock:
            if tnum > 0:
                self.w_stats.empty_runs = 0
            self.w_stats.queue = tnum
            self.w_stats.distribution = wkgroups
            self.w_stats.workers = len(all_workers)
        # I'll be greedy only if tasks assigned are equal to the limit
        # (meaning there could be others ready to be assigned)
        self.greedy = tnum >= limit
        logger.info("TICKER: workers are %s", len(all_workers))
        logger.info("TICKER: tasks are %s", tnum)

    def sleep(self):
        """Calculate the number of seconds to sleep."""
        time.sleep(self.w_stats.sleep)
        # should only sleep until next available task

    def set_worker_status(
        self,
        group_names=None,
        action=ACTIVE,
        exclude=None,
        limit=None,
        worker_name=None,
    ):
        """Internal function to set worker's status."""
        db = self.db
        ws = db.scheduler_worker
        if not group_names:
            group_names = self.group_names
        elif isinstance(group_names, str):
            group_names = [group_names]
        if worker_name:
            db(ws.worker_name == worker_name).update(status=action)
            return
        exclusion = exclude and exclude.append(action) or [action]
        if not limit:
            for group in group_names:
                db(
                    (ws.group_names.contains(group)) & (~ws.status.belongs(exclusion))
                ).update(status=action)
        else:
            for group in group_names:
                workers = db(
                    (ws.group_names.contains(group)) & (~ws.status.belongs(exclusion))
                )._select(ws.id, limitby=(0, limit))
                db(ws.id.belongs(workers)).update(status=action)

    def disable(self, group_names=None, limit=None, worker_name=None):
        """Set DISABLED on the workers processing `group_names` tasks.

        A DISABLED worker will be kept alive but it won't be able to process
        any waiting tasks, essentially putting it to sleep.
        By default, all group_names of Scheduler's instantation are selected
        """
        self.set_worker_status(
            group_names=group_names,
            action=DISABLED,
            exclude=[DISABLED, KILL, TERMINATE],
            limit=limit,
        )

    def resume(self, group_names=None, limit=None, worker_name=None):
        """Wakes a worker up (it will be able to process queued tasks)"""
        self.set_worker_status(
            group_names=group_names,
            action=ACTIVE,
            exclude=[KILL, TERMINATE],
            limit=limit,
        )

    def terminate(self, group_names=None, limit=None, worker_name=None):
        """Sets TERMINATE as worker status. The worker will wait for any
        currently running tasks to be executed and then it will exit gracefully
        """
        self.set_worker_status(
            group_names=group_names, action=TERMINATE, exclude=[KILL], limit=limit
        )

    def kill(self, group_names=None, limit=None, worker_name=None):
        """Sets KILL as worker status. The worker will be killed even if it's
        processing a task."""
        self.set_worker_status(group_names=group_names, action=KILL, limit=limit)

    def queue_task(self, function, pargs=[], pvars={}, **kwargs):
        """
        Queue tasks. This takes care of handling the validation of all
        parameters

        Args:
            function: the function (anything callable with a __name__)
            pargs: "raw" args to be passed to the function. Automatically
                jsonified.
            pvars: "raw" kwargs to be passed to the function. Automatically
                jsonified
            kwargs: all the parameters available (basically, every
                `scheduler_task` column). If args and vars are here, they
                should be jsonified already, and they will override pargs
                and pvars

        Returns:
            a dict just as a normal validate_and_insert(), plus a uuid key
            holding the uuid of the queued task. If validation is not passed
            ( i.e. some parameters are invalid) both id and uuid will be None,
            and you'll get an "error" dict holding the errors found.
        """
        if hasattr(function, "__name__"):
            function = function.__name__
        targs = "args" in kwargs and kwargs.pop("args") or dumps(pargs)
        tvars = "vars" in kwargs and kwargs.pop("vars") or dumps(pvars)
        tuuid = "uuid" in kwargs and kwargs.pop("uuid") or web2py_uuid()
        tname = "task_name" in kwargs and kwargs.pop("task_name") or function
        immediate = "immediate" in kwargs and kwargs.pop("immediate") or None
        cronline = kwargs.get("cronline")
        kwargs.update(
            function_name=function,
            task_name=tname,
            args=targs,
            vars=tvars,
            uuid=tuuid,
        )
        if cronline:
            try:
                start_time = kwargs.get("start_time", self.now)
                next_run_time = CronParser(cronline, start_time).next()
                kwargs.update(start_time=start_time, next_run_time=next_run_time)
            except Exception:
                pass
        if "start_time" in kwargs and "next_run_time" not in kwargs:
            kwargs.update(next_run_time=kwargs["start_time"])
        db = self.db
        rtn = db.scheduler_task.validate_and_insert(**kwargs)
        if not rtn.errors:
            rtn.uuid = tuuid
            if immediate:
                db((db.scheduler_worker.is_ticker == True)).update(status=PICK)
        else:
            rtn.uuid = None
        return rtn

    def task_status(self, ref, output=False):
        """
        Retrieves task status and optionally the result of the task

        Args:
            ref: can be

              - an integer : lookup will be done by scheduler_task.id
              - a string : lookup will be done by scheduler_task.uuid
              - a `Query` : lookup as you wish, e.g. ::

                    db.scheduler_task.task_name == 'test1'

            output(bool): if `True`, fetch also the scheduler_run record

        Returns:
            a single Row object, for the last queued task.
            If output == True, returns also the last scheduler_run record.
            The scheduler_run record is fetched by a left join, so it can
            have all fields == None

        """
        from pydal.objects import Query

        db = self.db
        sr = db.scheduler_run
        st = db.scheduler_task
        if isinstance(ref, integer_types):
            q = st.id == ref
        elif isinstance(ref, str):
            q = st.uuid == ref
        elif isinstance(ref, Query):
            q = ref
        else:
            raise SyntaxError("You can retrieve results only by id, uuid or Query")
        fields = [st.ALL]
        left = False
        orderby = ~st.id
        if output:
            fields = st.ALL, sr.ALL
            left = sr.on(sr.task_id == st.id)
            orderby = ~st.id | ~sr.id
        row = (
            db(q)
            .select(*fields, **dict(orderby=orderby, left=left, limitby=(0, 1)))
            .first()
        )
        if row and output:
            row.result = (
                row.scheduler_run.run_result
                and loads(row.scheduler_run.run_result, object_hook=_decode_dict)
                or None
            )
        return row

    def stop_task(self, ref):
        """Shortcut for task termination.

        If the task is RUNNING it will terminate it, meaning that status
        will be set as FAILED.

        If the task is QUEUED, its stop_time will be set as to "now",
            the enabled flag will be set to False, and the status to STOPPED

        Args:
            ref: can be

              - an integer : lookup will be done by scheduler_task.id
              - a string : lookup will be done by scheduler_task.uuid

        Returns:
            - 1 if task was stopped (meaning an update has been done)
            - None if task was not found, or if task was not RUNNING or QUEUED

        Note:
            Experimental
        """
        db = self.db
        st = db.scheduler_task
        sw = db.scheduler_worker
        if isinstance(ref, integer_types):
            q = st.id == ref
        elif isinstance(ref, str):
            q = st.uuid == ref
        else:
            raise SyntaxError("You can retrieve results only by id or uuid")
        task = db(q).select(st.id, st.status, st.assigned_worker_name)
        task = task.first()
        rtn = None
        if not task:
            return rtn
        if task.status == "RUNNING":
            q = sw.worker_name == task.assigned_worker_name
            rtn = db(q).update(status=STOP_TASK)
        elif task.status == "QUEUED":
            rtn = db(q).update(stop_time=self.now(), enabled=False, status=STOPPED)
        return rtn

    def get_workers(self, only_ticker=False):
        """Returns a dict holding `worker_name : {**columns}`
        representing all "registered" workers
        only_ticker returns only the workers running as a TICKER,
        if there are any
        """
        db = self.db
        if only_ticker:
            workers = db(db.scheduler_worker.is_ticker == True).select()
        else:
            workers = db(db.scheduler_worker.id).select()
        all_workers = {}
        for row in workers:
            all_workers[row.worker_name] = Storage(
                status=row.status,
                first_heartbeat=row.first_heartbeat,
                last_heartbeat=row.last_heartbeat,
                group_names=row.group_names,
                is_ticker=row.is_ticker,
                worker_stats=row.worker_stats,
            )
        return all_workers


def main():
    """
    allows to run worker without python web2py.py .... by simply::

        python gluon/scheduler.py

    """
    import optparse

    parser = optparse.OptionParser()
    parser.add_option(
        "-w",
        "--worker_name",
        dest="worker_name",
        default=None,
        help="start a worker with name",
    )
    parser.add_option(
        "-b",
        "--heartbeat",
        dest="heartbeat",
        default=10,
        type="int",
        help="heartbeat time in seconds (default 10)",
    )
    parser.add_option(
        "-L",
        "--logger_level",
        dest="logger_level",
        default=30,
        type="int",
        help="set debug output level (0-100, 0 means all, 100 means none;default is 30)",
    )
    parser.add_option(
        "-E",
        "--empty-runs",
        dest="max_empty_runs",
        type="int",
        default=0,
        help="max loops with no grabbed tasks permitted (0 for never check)",
    )
    parser.add_option(
        "-g",
        "--group_names",
        dest="group_names",
        default="main",
        help="comma separated list of groups to be picked by the worker",
    )
    parser.add_option(
        "-f",
        "--db_folder",
        dest="db_folder",
        default="/Users/mdipierro/web2py/applications/scheduler/databases",
        help="location of the dal database folder",
    )
    parser.add_option(
        "-u",
        "--db_uri",
        dest="db_uri",
        default="sqlite://storage.sqlite",
        help="database URI string (web2py DAL syntax)",
    )
    parser.add_option(
        "-t",
        "--tasks",
        dest="tasks",
        default=None,
        help="file containing task files, must define"
        + "tasks = {'task_name':(lambda: 'output')} or similar set of tasks",
    )
    parser.add_option(
        "-U",
        "--utc-time",
        dest="utc_time",
        default=False,
        help="work with UTC timestamps",
    )
    (options, args) = parser.parse_args()
    if not options.tasks or not options.db_uri:
        print(USAGE)
    if options.tasks:
        path, filename = os.path.split(options.tasks)
        if filename.endswith(".py"):
            filename = filename[:-3]
        sys.path.append(path)
        print("importing tasks...")
        tasks = __import__(filename, globals(), locals(), [], -1).tasks
        print("tasks found: " + ", ".join(list(tasks.keys())))
    else:
        tasks = {}
    group_names = [x.strip() for x in options.group_names.split(",")]

    logging.getLogger().setLevel(options.logger_level)

    print("groups for this worker: " + ", ".join(group_names))
    print("connecting to database in folder: " + options.db_folder or "./")
    print("using URI: " + options.db_uri)
    db = DAL(options.db_uri, folder=options.db_folder, decode_credentials=True)
    print("instantiating scheduler...")
    scheduler = Scheduler(
        db=db,
        worker_name=options.worker_name,
        tasks=tasks,
        migrate=True,
        group_names=group_names,
        heartbeat=options.heartbeat,
        max_empty_runs=options.max_empty_runs,
        utc_time=options.utc_time,
    )
    signal.signal(signal.SIGTERM, lambda signum, stack_frame: sys.exit(1))
    print("starting main worker loop...")
    scheduler.loop()


if __name__ == "__main__":
    main()



import uuid
import hashlib
import jwt

from py4web.core import request, abort, DAL, Field


class OAuthServer(object):

    algorithms = ["HS256", "RS256"]

    def __init__(self, auth, secret):
        self.secret = secret
        self.auth = auth
        self.define_tables()

    def define_tables(self):
        db = self.auth.db
        db.define_table(
            "oauth2",
            Field("registrant_id", "reference auth_user"),
            Field("client_secret"),
        )
        db.commit()

    def register_new_client(self, registrant_id):
        client_secret = str(uuid.uuid4())
        self.auth.db.oauth2.insert(
            registrant_id=registrant_id, client_secret=client_secret
        )

    def handle_request(self, auth, path, get_vars, post_vars):
        if path == "login":
            # WIP
            client_id = request.get("client_id")
            info = {}  # retrieve the user info upon login
            code = jwt.encode(info, self.secret + client_id, algorithm="HS256")
        elif path == "callback":
            db = self.auth.db
            code = (get_vars.get("code"),)
            client_id = (get_vars.get("client_id"),)
            client_secret = (get_vars.get("client_secret"),)
            redirect_uri = (get_vars.get("callback_url"),)
            grant_type = get_vars.get("grant_type")
            if (
                not grant_type == "authorization_code"
                or not hashlib.sha1(client_secret).hexdigest() == client_id
                or not db(db.oauth2.client_secret == client_secret).count()
            ):
                abort(404)
            info = jwt.decode(code, self.secret + client_id, algorithms=["HS256"])
            access_token = jwt.encode(info, self.secret, algorithm="HS256")
            return dict(access_token=access_token)
        elif path == "profile":
            access_token = request.environ.get("HTTP_AUTHORIZATION", "")[7:]
            info = jwt.decode(access_token, self.secret, algorithms=self.algorithms)
            return info
        else:
            abort(404)
